/*
 $Revision: 1.14 $
 $Date: 2009/12/03 13:46:01 $
 $Author: apollan $
 $Name:  $
 $Id: ace.c,v 1.14 2009/12/03 13:46:01 apollan Exp $
 */

/*-------------------------------------------------------------------*/
/* ACE.C      Antti Pollanen                                         */
/*                                                                   */
/* Console interface fo ACE -algorithm                               */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "cbace"
#define  VersionNumber  "Version 0.23"
#define  LastUpdated    "01.12.2009" /* AP */
#define  MAXFILENAME    64
#define  FACTFILE       "cbace.fac"

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* AP   0.23  Multiple partitions bug fix                            */
/* AP   0.22  Better structure                                       */
/* AP   0.21  CTK MergeCluster fixed                                 */
/* AP   0.20  Error in MergeClusters function fixed                  */
/* AP   0.11  Stardardization for coding and commenting              */
/* AP   0.10  Bug fix: CostComparison returned wrong value           */
/* AP   0.02  Many changes, algorithm not working yet                */
/* AP   0.01  Initial implementation                                 */
/*                                                                   */
/* ----------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "parametr.h"
#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "heap.h"
#include "stack.h"
#include "math.h"

#include "ctk.h"
#include "ace.h"


/* ========================== FUNCTIONS ============================ */

int IsInHeap(HEAP* pHeap, CLUSTERPAIR* pPair)
  {
  if (pPair->Index > 0 && pPair->Index <= Heap_size(pHeap))
    {
    return 1;
    }

  return 0;
  } /* IsInHeap */

/*-------------------------------------------------------------------*/

int CostComparision(void* elem1, void* elem2, void* info)
  {
  CLUSTERPAIR* c1 = (CLUSTERPAIR*) elem1;
  CLUSTERPAIR* c2 = (CLUSTERPAIR*) elem2;

  if (c1->Cost < c2->Cost)
    {
    return -1;
    }
  else if (c2->Cost < c1->Cost)
    {
    return 1;
    }

  return -1;
  } /* CostComparision */

/*-------------------------------------------------------------------*/

// Check if this gives correct results
double VectorMergeCost(TRAININGSET* pTS, int v1, int v2)
  {
  double error = 0.0;
  int j;

  for (j = 0; j < VectorSize(pTS); j++)
    {
    if (VectorScalar(pTS, v1, j) != VectorScalar(pTS, v2, j))
      {
      error += 0.69314718055994529; // error -= log(0.5);
      }
    }

  return 2 * error;
  } /* VectorMergeCost */

/*-------------------------------------------------------------------*/

double MergeCost(CATCODEBOOK* pCCB, int c1, int c2)
  {
  int j, k;
  int count;
  int occurrences;
  double error = 0.0;
  double probability;

  count = Count(pCCB, c1) + Count(pCCB, c2);

  for (j = 0; j < VectorSize(pCCB->pCodebook); j++)
    {
    for (k = 0; k < Categories(pCCB, j); k++)
      {
      occurrences = CategoryInstances(pCCB, c1, j, k) +
                    CategoryInstances(pCCB, c2, j, k);

      if (occurrences != 0)
        {
        probability = (double) occurrences / (double) count;
        error -= probability * log(probability);
        }
      }
    }


  double newError = count * error;
  double oldError1 = CATClusterEntropy(pCCB, c1);
  double oldError2 = CATClusterEntropy(pCCB, c2);
  return newError - (oldError1 + oldError2);
  } /* MergeCost */

/*-------------------------------------------------------------------*/

int ACEInit(PARTITIONING* pP,
            TRAININGSET* pTS,
            ACEDATA* data,
            int quietLevel)
  {
  int i, j;
  int BookSize = BookSize(pTS);
  int PairIndex;

  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->heap);
  CATCODEBOOK* pCCB = &(data->CCB);
  CODEBOOK* pCB = &(data->CB);

  CATPrint(quietLevel, INFOMESSAGE, "Initializing...\n");

  InitRandom();
  UTMCreate(pUtm, BookSize);
  Heap_init(pHeap, pUtm->Size, CostComparision);
  data->Pairs = (CLUSTERPAIR*) allocate(sizeof(CLUSTERPAIR) * pUtm->Size);
  CLUSTERPAIR* Pairs = data->Pairs;

  CreateNewCodebook(pCB, BookSize, pTS);
  if (CATCreateNewCodebook(pCCB, pCB, pTS))
    {
    CATPrint(quietLevel, ERRORMESSAGE, "Allocating CCB failed!\n");
    return 1;
    }

  CreateNewPartitioning(pP, pTS, BookSize);
  for (i = 0; i < BookSize; i++)
    {
    CATChangePartition(pCCB, pTS, pP, -1, i, i);
    }

  PairIndex = 0;
  for (i = 0; i < BookSize; i++)
    {
    for (j = i + 1; j < BookSize; j++)
      {
      Pairs[PairIndex].Cluster1 = i;
      Pairs[PairIndex].Cluster2 = j;
      Pairs[PairIndex].Cost = VectorMergeCost(pTS, i, j);

      UTMSet(pUtm, i, j, &(Pairs[PairIndex]));
      Heap_insert(pHeap, &(Pairs[PairIndex]),
                  NULL, &(Pairs[PairIndex].Index));
      PairIndex++;
      }

    CATPrintV(quietLevel, DEBUGMESSAGE, "Init %d\n", i);
    }

  data->x = BookSize;
  CATPrint(quietLevel, INFOMESSAGE, "Initializing completed.\n");

  return 0;
  } /* ACEInit */

/*-------------------------------------------------------------------*/

void RenameCluster(PARTITIONING* pP, TRAININGSET* pTS, ACEDATA* data,
    int quietLevel, int c2)
  {
  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->heap);
  CLUSTERPAIR* Pairs = data->Pairs;
  int x = data->x;

  int i;
  int last = x - 1;
  int renamed = 0;

  // Rename last cluster to c2.
  if (c2 != last)
    {
    CATPrintV(quietLevel, INFOMESSAGE, "Renaming %d to %d.\n", last, c2);

    for (i = 0; i < last; i++)
      {
      if (c2 != i)
        {
        // Pair (c2, last) is removed already.
        CLUSTERPAIR* MovePair = UTMGet(pUtm, i, last);

        if (MovePair->Cluster1 != i && MovePair->Cluster2 != i)
          {
          CATPrintV(quietLevel, ERRORMESSAGE,
                    "ERROR: renaming wrong pair (%d, %d).\n",
                    MovePair->Cluster1, MovePair->Cluster2);
          }

        renamed++;

        if (c2 < i)
          {
          MovePair->Cluster1 = c2;
          MovePair->Cluster2 = i;
          UTMSet(pUtm, c2, i, MovePair);
          }
        else
          {
          MovePair->Cluster1 = i;
          MovePair->Cluster2 = c2;
          UTMSet(pUtm, i, c2, MovePair);
          }
        }
      }
    }

  CATPrintV(quietLevel, DEBUGMESSAGE,
            "Renamed %d elements.\n", renamed);

  if (quietLevel == DEBUGMESSAGE)
    {
    for (i = 0; i < pUtm->Size; i++)
      {
      if (Pairs[i].Index > 0 && Pairs[i].Index <= Heap_size(pHeap)
          && (Pairs[i].Cluster1 >= last || Pairs[i].Cluster2 >= last))
        {
        CATPrintV(quietLevel, ERRORMESSAGE,
                  "ERROR: Pair (%d, %d) found!\n",
                  Pairs[i].Cluster1, Pairs[i].Cluster2);
        }
      }
    }

  } /* RenameCluster */

/*-------------------------------------------------------------------*/

void MergeClusters(PARTITIONING* pP, TRAININGSET* pTS, ACEDATA* data,
    int quietLevel, int c1, int c2)
  {
  CATPrintV(quietLevel, INFOMESSAGE, "Merging %d and %d\n", c1, c2);

  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->heap);
  CATCODEBOOK* pCCB = &(data->CCB);
  int x = data->x;

  int i;
  int removed = 0;
  CLUSTERPAIR *OldPair;

  // Remove cluster c2.
  for (i = 0; i < x; i++)
    {
    // Pair(x, x) doesn't exist and Pair(c1, c2) is already removed.
    if (i != c2 && i != c1)
      {
      OldPair = UTMGet(pUtm, c2, i);

      // This should be unnecessary.
      if (IsInHeap(pHeap, OldPair))
        {
        Heap_remove(pHeap, OldPair->Index, NULL);

        if (IsInHeap(pHeap, OldPair))
          {
          CATPrintV(quietLevel, ERRORMESSAGE,
                    "ERROR: heap is broken (index %d)!\n",
                    OldPair->Index);
          }

        if (OldPair->Cluster1 != c2 && OldPair->Cluster2 != c2)
          {
          CATPrintV(quietLevel, ERRORMESSAGE,
                    "ERROR: removing wrong pair (%d, %d).\n",
                    OldPair->Cluster1, OldPair->Cluster2);
          }

        removed++;
        }
      else
        {
        CATPrintV(quietLevel, ERRORMESSAGE,
                  "Cluster pair %d, %d doesn't exist.\n", c2, i);
        }
      }
    }

  CATPrintV(quietLevel, DEBUGMESSAGE, "Removed %d pairs.\n", removed);
  CATMergeClusters(pCCB, pTS, pP, c1, c2);
  } /* MergeClusters */

/*-------------------------------------------------------------------*/

void UpdateMergeCosts(PARTITIONING* pP, TRAININGSET* pTS, ACEDATA* data,
    int quietLevel, int c1)
  {
  CATPrint(quietLevel, INFOMESSAGE, "Updating merge costs\n");

  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->heap);
  CATCODEBOOK* pCCB = &(data->CCB);
  int x = data->x;

  CLUSTERPAIR *OldPair;
  int i;
  int last = x - 1;

  for (i = 0; i < last; i++)
    {
    if (i != c1)
      {
      OldPair = UTMGet(pUtm, c1, i);

      if (IsInHeap(pHeap, OldPair))
        {
        OldPair->Cost = MergeCost(pCCB, c1, i);
        Heap_update(pHeap, OldPair->Index, NULL);
        }
      else
        {
        PrintMessage("Heap size %d\n", Heap_size(pHeap));
        CATPrintV(quietLevel, ERRORMESSAGE,
                  "Cluster pair %d, %d doesn't exist.\n", c1, i);
        }
      }
    }

  } /* UpdateMergeCosts */

/*-------------------------------------------------------------------*/

CLUSTERPAIR* SelectPair(HEAP* pHeap)
  {
  CLUSTERPAIR* pPair;
  CLUSTERPAIR* pFirst;
  CLUSTERPAIR* pSelected;
  STACK* stack = S_make();
  int size = 0;

  pFirst = (CLUSTERPAIR*) Heap_removeroot(pHeap, NULL);
  S_push(stack, pFirst);
  size++;
  int selected;
  int tie = 1;

  while (tie && !Heap_isempty(pHeap))
    {
    pPair = (CLUSTERPAIR*) Heap_removeroot(pHeap, NULL);

    if (fabs(pPair->Cost - pFirst->Cost) < EPSILON)
      {
      S_push(stack, pPair);
      size++;
      }
    else
      {
      Heap_insert(pHeap, pPair, NULL, &(pPair->Index));
      tie = 0;
      }
    }

  selected = IRZ(size);

  while (!S_empty(stack))
    {
    pPair = (CLUSTERPAIR*)S_pop(stack);
    size--;

    if (size == selected)
      {
      pSelected = pPair;
      }
    else
      {
      Heap_insert(pHeap, pPair, NULL, &(pPair->Index));
      }
    }

  S_free(stack);

  return pSelected;
  } /* SelectPair */

/*-------------------------------------------------------------------*/

int ACEIteration(PARTITIONING* pP, TRAININGSET* pTS, ACEDATA* data,
    int quietLevel)
  {
  HEAP* pHeap = &(data->heap);
  int x = data->x;

  CLUSTERPAIR* pPair;
  int c1, c2;

  CATPrintV(quietLevel, INFOMESSAGE, "--- Iteration starts ---\n", x);

  pPair = SelectPair(pHeap);
  c1 = pPair->Cluster1;
  c2 = pPair->Cluster2;

  MergeClusters(pP, pTS, data, quietLevel, c1, c2);
  RenameCluster(pP, pTS, data, quietLevel, c2);
  UpdateMergeCosts(pP, pTS, data, quietLevel, c1);

  data->x = --x;
  CATPrintV(quietLevel, INFOMESSAGE, "--- Codebook size %d ---\n", x);
  CATPrintV(quietLevel, INFOMESSAGE, "--- Iteration ends ---\n", x);

  return 0;
  } /* ACEIteration */

/*-------------------------------------------------------------------*/

int ACEFree(ACEDATA* data, int quietLevel)
  {
  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->heap);
  CATCODEBOOK* pCCB = &(data->CCB);
  CLUSTERPAIR* Pairs = data->Pairs;

  // Deallocate resources.
  FreeCodebook(pCCB->pCodebook);
  CATFreeCodebook(pCCB);
  UTMFree(pUtm);
  Heap_free(pHeap);
  deallocate(Pairs);

  return 0;
  } /* ACEFree */

/*-------------------------------------------------------------------*/

int PerformACE(PARTITIONING* pP,
               TRAININGSET* pTS,
               int k,
               int quietLevel)
  {
  int i, error;
  int BookSize = BookSize(pTS);
  ACEDATA data;

  CATPrint(quietLevel, INFOMESSAGE, "Clustering...\n");
  error = ACEInit(pP, pTS, &data, quietLevel);

  for (i = BookSize; i > k && !error; i--)
    {
    ACEIteration(pP, pTS, &data, quietLevel);
    }

  error = error || ACEFree(&data, quietLevel);
  CATPrint(quietLevel, INFOMESSAGE, "Clustering completed.\n");

  return error;
  } /* PerformACE */

