/*
$Revision: 1.9 $
$Date: 2009/12/03 13:46:01 $
$Author: apollan $
$Name:  $
$Id: cbace.c,v 1.9 2009/12/03 13:46:01 apollan Exp $
*/

/*-------------------------------------------------------------------*/
/* ACE.C      Antti Pollanen                                         */
/*                                                                   */
/* Console interface for ACE -algorithm                               */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "cbace"
#define  VersionNumber  "Version 0.23"
#define  LastUpdated    "01.12.2009" /* AP */
#define  MAXFILENAME    64
#define  FACTFILE       "cbace.fac"

/*-------------------------------------------------------------------*/

#include <stdio.h>
#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "heap.h"

#include "ctk.h"
#include "ace.h"

/* Constants used in parameter mapping are defined in chm.h. */
#include "parametr.c"


/* ========================== PROTOTYPES =========================== */

void PrintInitialData(int quietLevel, char *TSName, char *PAName,
  int k, int maxK);

void PrintInfo(void);

int main(int argc, char* argv[]);


/* ========================== FUNCTIONS ============================ */

void PrintInitialData(int quietLevel, char *TSName, char *PAName,
  int k, int maxK)
  {
  if (quietLevel >= INFOMESSAGE)
    {
    PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
    PrintMessage("Training Set\t\t = %s\n", TSName);
    PrintMessage("Result Partitioning\t = %s\n", PAName);

    if (maxK > k)
      {
      PrintMessage("Codebook size\t\t = %d - %d\n", k, maxK);
      }
    else
      {
      PrintMessage("Codebook size\t\t = %d\n", k);
      }
    }
  }  /* PrintInitialData() */

/*-------------------------------------------------------------------*/

void PrintInfo(void)
  {
  PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
  PrintMessage("Console interface for the hierarcical methods for categorical data.\n");
  PrintMessage("Usage: %s [%coption] ", ProgName, OPTION_SYMBOL);
  PrintMessage("<training set> <result partitioning>\n");
  PrintMessage("For example: %s bridge tmp\n", ProgName);
  PrintMessage("\n  Options:\n");
  PrintOptions();
  PrintMessage("\n");
  }  /* PrintInfo() */

/*-------------------------------------------------------------------*/

int PerformACEMultipleResults(PARTITIONING* pP,
                              TRAININGSET* pTS,
                              int k,
                              int maxK,
                              int quietLevel,
                              char* PAName)
  {
  char    RealPAName[MAXFILENAME + 10];
  int     error;
  int     i;
  ACEDATA data;

  error = ACEInit(pP, pTS, &data, quietLevel);

  if (BookSize(pTS) <= maxK)
    {
    sprintf(RealPAName, "%.*s%d.pa", strlen(PAName) - 3, PAName, BookSize(pTS));
    WritePartitioning(RealPAName, pP, pTS, 1);
    }

  for (i = BookSize(pTS); i > k && !error; i--)
    {
    error = ACEIteration(pP, pTS, &data, quietLevel);

    if (i-1 <= maxK)
      {
      sprintf(RealPAName, "%.*s%d.pa", strlen(PAName) - 3, PAName, i-1);
      WritePartitioning(RealPAName, pP, pTS, 1);
      }
    }

  ACEFree(&data, quietLevel);
  return error;
  }

/*-------------------------------------------------------------------*/

int main(int argc, char* argv[])
  {
  TRAININGSET   TS;
  PARTITIONING  P;
  char          TSName[MAXFILENAME] = {'\0'};
  char          PAName[MAXFILENAME] = {'\0'};
  ParameterInfo paraminfo[2] = { { TSName,  FormatNameTS, 0, INFILE },
                                 { PAName,  FormatNamePA, 0, OUTFILE } };
  int error;
  ParseParameters(argc, argv, 2, paraminfo);

  /* training set doesn't exist */
  if (!ExistFile(TSName))
    {
    ErrorMessage("\nERROR: File doesn't exist: %s\n\n", TSName);
    ExitProcessing(FATAL_ERROR);
    }

  /* result codebook exists and we are told not to overwrite */
  if (ExistFile(PAName) && !Value(OverWrite))
    {
    ErrorMessage("\nERROR: File already exists: %s\n\n", PAName);
    ExitProcessing(FATAL_ERROR);
    }

  ReadTrainingSet(TSName, &TS);

  /* codebook cannot contain more vectors than training set */
  if (BookSize(&TS) < Value(Clusters))
    {
    ErrorMessage("\nERROR: Number of vectors in training set ");
    ErrorMessage("(%d) < number of clusters ", BookSize(&TS));
    ErrorMessage("(%d)!\n\n", Value(Clusters));
    FreeCodebook(&TS);
    ExitProcessing(FATAL_ERROR);
    }

  PrintInitialData(Value(QuietLevel), TSName, PAName, Value(Clusters), Value(MaxClusters));
  CreateNewPartitioning(&P, &TS, Value(Clusters));

  error = PerformACEMultipleResults(&P,
                                    &TS,
                                    Value(Clusters),
                                    Value(MaxClusters),
                                    Value(QuietLevel),
                                    PAName);

  if (error)
    {
    ErrorMessage("ERROR: Clustering failed!\n");
    FreeCodebook(&TS);
    FreePartitioning(&P);
    ExitProcessing(FATAL_ERROR);
    }
  else
    {
    WritePartitioning(PAName, &P, &TS, Value(OverWrite));
    }

  FreeCodebook(&TS);
  FreePartitioning(&P);

  checkmemory();

  return 0;
  } /* main */
