/*
$Revision: 1.12 $
$Date: 2009/12/03 13:46:01 $
$Author: apollan $
$Name:  $
$Id: ceval.c,v 1.12 2009/12/03 13:46:01 apollan Exp $
*/

/*-------------------------------------------------------------------*/
/* CEVAL.C      Antti Pollanen                                        */
/*                                                                   */
/* Evaluation module for categorical clustering                      */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName        "CEVAL"
#define VersionNumber   "0.33"
#define LastUpdated     "15.10.2009" /* AP */

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* AP   0.33  Classification error improved. QuietLevel added.       */
/* AP   0.32  Dataset entropy added to entropy criterion             */
/* AP   0.31  Methods verified                                       */
/* AP   0.30  Classification error implemented                       */
/* AP   0.21  Stardardization for coding and commenting              */
/* AP   0.20  Automated testing scripts moved to own module          */
/* AP   0.10  Automated testing script added                         */
/* AP   0.01  Initial implementation                                 */
/*                                                                   */
/* ----------------------------------------------------------------- */

#include "cb.h"

#include "memctrl.h"
#include "interfc.h"
#include "ctk.h"
#include "ceval.h"

#include "math.h"


/* ========================== PROTOTYPES =========================== */

double CATErrorEntropy        (CATCODEBOOK* pCCB,
                               PARTITIONING* pP,
                               TRAININGSET *pTS,
                               int quietLevel);

double CATErrorCategoryUtility(CATCODEBOOK* pCCB,
                               PARTITIONING* pP,
                               TRAININGSET *pTS,
                               int quietLevel);

double CATErrorClassification (CATCODEBOOK* pCCB,
                               PARTITIONING* pP,
                               TRAININGSET *pTS,
                               int quietLevel);

/* ========================== FUNCTIONS ============================ */

double CATError(CATERRORTYPE errortype,
                CATCODEBOOK* pCCB,
                PARTITIONING* pP,
                TRAININGSET *pTS,
                int quietLevel)
  {
  switch (errortype)
    {
    case ENTROPYERROR:
      return CATErrorEntropy(pCCB, pP, pTS, quietLevel);
      break;
    case CATEGORYUTILITYERROR:
      return CATErrorCategoryUtility(pCCB, pP, pTS, quietLevel);
      break;
    case CLASSIFICATIONERROR:
      return CATErrorClassification(pCCB, pP, pTS, quietLevel);
      break;
    }

  return UNKNOWNTYPE;
  } /* CATError */

/* ========================== ENTROPY ============================== */

double CATErrorEntropy(CATCODEBOOK* pCCB,
                       PARTITIONING* pP,
                       TRAININGSET *pTS,
                       int quietLevel)
  {
  int i;
  double total = 0.0;
  CODEBOOK *pCB = pCCB->pCodebook;
  CATGenerateOptimalCodebook(pP, pTS, pCCB, DISTRIBUTIONREPR);

  for (i = 0; i < BookSize(pCB); i++)
    {
    total += CATClusterEntropy(pCCB, i);
    }

  total = (double)total / (double)BookSize(pTS);
  return total;
  } /* CATEntropyError() */

/* ====================== CATEGORY UTILITY ========================= */

double CATErrorCategoryUtility(CATCODEBOOK* pCCB,
                               PARTITIONING* pP,
                               TRAININGSET *pTS,
                               int quietLevel)
  {
  int i;            // attributes A.
  int c;            // clusters C.
  int j;            // possible values D.
  double total = 0; // CU
  double sum;
  CODEBOOK* pCB = pCCB->pCodebook;

  int t;
  double catins;    // Number of instances in the category
  double pc;        // Probability of the category in the current cluster.
  double pa;        // Probability of the category in all clusters.

  CATGenerateOptimalCodebook(pP, pTS, pCCB, DISTRIBUTIONREPR);

  for (c = 0; c < BookSize(pCB); c++)
    {
	  if (Count(pCCB, c) != 0)
	    {
      sum = 0;
      for (i = 0; i < VectorSize(pCB); i++)
        {
        for (j = 0; j < Categories(pCCB, i); j++)
          {
          catins = 0;

          for (t = 0; t < BookSize(pCB); t++)
            {
            catins += CategoryInstances(pCCB, t, i, j);
            }

          pa = (double)catins / (double)BookSize(pTS);
          pc = (double)CategoryInstances(pCCB, c, i, j) /
               (double)Count(pCCB, c);
          sum += (pow(pc, 2) - pow(pa, 2));
          }
        }

      total += ((double)Count(pCCB, c) / BookSize(pTS)) * sum;
      }
	  }

  return total;
  } /* CATErrorCategoryUtility() */

/* ====================== CLASSIFICATION =========================== */

void CountAccuracyTable(CATCODEBOOK* pCCB,
                        PARTITIONING* pP,
                        TRAININGSET *pTS,
                        int* correct)
  {

  int c, v, i;
  int vectorClass, vectorCluster;
  int classOffset = -1;

  // Find out first indexing number (0/1).
  for (i = 0; i < BookSize(pTS); i++)
    {
    if (VectorScalar(pTS, i, 0) == 0)
      {
      classOffset = 0;
      }
    }

  for (c = 0; c < BookSize(pCCB->pCodebook); c++)
    {
    for (v = 0; v < BookSize(pCCB->pCodebook); v++)
      {
      for (i = 0; i < BookSize(pTS); i++)
        {
        vectorClass   = VectorScalar(pTS, i, 0) + classOffset;
        vectorCluster = Map(pP, i);

        if (vectorClass == c && vectorCluster == v)
          {
          correct[c*BookSize(pCCB->pCodebook) + v]++;
          }
        }
      }
    }
  } /* CountAccuracyTable */

/*-------------------------------------------------------------------*/

void PrintAccuracyTable(int* correct, int bookSize, int quietLevel)
  {
  int i, j;

  CATPrint(quietLevel, DEBUGMESSAGE, "Accuracy Table\n");

  for (i = 0; i < bookSize; i++)
    {
    for (j = 0; j < bookSize; j++)
      {
      CATPrintV(quietLevel, DEBUGMESSAGE, "%d ", correct[i*bookSize + j]);
      }

    CATPrint(quietLevel, DEBUGMESSAGE, "\n");
    }

  CATPrint(quietLevel, DEBUGMESSAGE, "\n");
  } /* PrintAccuracyTable */

/*-------------------------------------------------------------------*/

void PrintPermutationAccuracy(int* mapping, int bookSize, int accuracy, int quietLevel)
  {
  int j;
  for (j = 0; j < bookSize; j++)
    {
    CATPrintV(quietLevel, DEBUGMESSAGE, "%d ", mapping[j]);
    }

  CATPrintV(quietLevel, DEBUGMESSAGE, " accuracy %d\n", accuracy);
  } /* PrintPermutationAccuracy */

/*-------------------------------------------------------------------*/

int CountMappingAccuracy(int* correct, int bookSize, int* mapping)
  {
  int i, accuracy = 0;

  for (i = 0; i < bookSize; i++)
    {
    accuracy += correct[mapping[i]*bookSize + i];
    }

  return accuracy;
  } /* CountMappingAccuracy */

/*-------------------------------------------------------------------*/

double CATErrorClassification(CATCODEBOOK* pCCB,
                              PARTITIONING* pP,
                              TRAININGSET *pTS,
                              int quietLevel)
  {
  CODEBOOK* pCB = pCCB->pCodebook;
  int bookSize = BookSize(pCB);
  int i;
  double e, r;
  int* correct;
  int* mapping;
  int accuracy, bestAccuracy;

  bestAccuracy = -1;
  mapping = (int*)allocate(sizeof(int)*bookSize);
  correct = (int*)allocate(sizeof(int)*bookSize*bookSize);

  for (i = 0; i < bookSize*bookSize; i++)
    {
    correct[i] = 0;
    }

  CATScaleTrainingSet(pTS, pTS);
  CountAccuracyTable(pCCB, pP, pTS, correct);
  PrintAccuracyTable(correct, bookSize, quietLevel);

  for (i = 0; i < CATFactorial(bookSize); i++)
    {
    CATPermutation(mapping, bookSize, i);
    accuracy = CountMappingAccuracy(correct, bookSize, mapping);
    PrintPermutationAccuracy(mapping, bookSize, accuracy, quietLevel);

    if (accuracy > bestAccuracy)
      {
      bestAccuracy = accuracy;
      }
    }

  CATPrint(quietLevel, DEBUGMESSAGE, "\n");

  r = (double)bestAccuracy / (double)BookSize(pTS);
  e = 1 - r;

  return e;
  } /* CATHuangClassificationError() */

/*-------------------------------------------------------------------*/

void RemoveOutliers(TRAININGSET *pTS,
                    PARTITIONING* pP,
                    PARTITIONING* pNewP,
                    int outlierCluster)
  {
  int i, used = 0, map;

  int oldSize = BookSize(pTS);
  int newSize = BookSize(pTS) - CCFreq(pP, outlierCluster);
  DecreaseCodebookSize(pTS, newSize);
  CreateNewPartitioning(pNewP, pTS, PartitionCount(pP) - 1);

  for (i = 0; i < oldSize; i++)
    {
    if (Map(pP, i) != outlierCluster)
      {
      CopyVector(Vector(pTS, i), Vector(pTS, used), VectorSize(pTS));
      map = (Map(pP, i) > outlierCluster ? Map(pP, i) - 1 : Map(pP, i));
      ChangePartition(pTS, pNewP, map, used);
      used++;
      }
    }
  } /* RemoveOutliers */
