/*
$Revision: 1.15 $
$Date: 2010/02/28 11:57:46 $
$Author: apollan $
$Name:  $
$Id: cbckm.c,v 1.15 2010/02/28 11:57:46 apollan Exp $
*/

/*-------------------------------------------------------------------*/
/* CBCKM.C     Antti Pollanen                                        */
/*                                                                   */
/* Console interface for categorical k-means algoritms.              */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "cbckm"
#define  VersionNumber  "Version 0.40 (CTK 0.61)"
#define  LastUpdated    "28.02.2010" /* AP */
#define  MAXFILENAME    64
#define  FACTFILE       "cbckm.fac"

/*-------------------------------------------------------------------*/

#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"

#include "ctk.h"
#include "ckm.h"

#include "math.h"
#include "parametr.c"


/* ========================== PROTOTYPES =========================== */

void PrintInitialData(int quietLevel,
                      char *TSName,
                      char *PAName,
                      int clus,
                      int kmIter,
                      CLUSTERINGMETHOD method,
                      CATINITMETHOD initMethod);

void PrintInfo(void);

int main(int argc, char* argv[]);


/* ========================== FUNCTIONS ============================ */

void PrintInitialData(int quietLevel,
                      char *TSName,
                      char *PAName,
                      int clus,
                      int kmIter,
                      CLUSTERINGMETHOD method,
                      CATINITMETHOD initMethod)
  {
  if (quietLevel > 1)
    {
    PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
    PrintMessage("Training Set\t\t = %s\n", TSName);
    PrintMessage("Result Partitioning\t = %s\n", PAName);
    PrintMessage("Codebook size\t\t = %d\n", clus);
    PrintMessage("K-means iterations\t = %d\n", kmIter);
    PrintMessage("Clustering method\t = %d\n", method);
    PrintMessage("Init method\t\t = %d\n", initMethod);
    PrintMessage("\n");
    }
  }  /* PrintInitialData() */

/*-------------------------------------------------------------------*/

void PrintInfo(void)
  {
  PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
  PrintMessage("Console interface for the categorical k-means.\n");
  PrintMessage("Usage: %s [%coption] ", ProgName, OPTION_SYMBOL);
  PrintMessage("<training set> <result partitioning>\n");
  PrintMessage("For example: %s bridge tmp\n", ProgName);
  PrintMessage("\n  Options:\n");
  PrintOptions();
  PrintMessage("\n");
  }  /* PrintInfo() */

/*-------------------------------------------------------------------*/

int main(int argc, char* argv[])
  {
  TRAININGSET   TS;
  CODEBOOK      CB;
  CATCODEBOOK   CCB;
  PARTITIONING  P;
  char          TSName[MAXFILENAME] = {'\0'};
  char          PAName[MAXFILENAME] = {'\0'};
  ParameterInfo paraminfo[2] = { { TSName,  FormatNameTS, 0, INFILE },
                                 { PAName,  FormatNamePA, 0, OUTFILE } };
  int result;
  CATDISTANCETYPE distanceMetric;
  CATCENTROIDTYPE representativeType;
  CATINITMETHOD   initMethod = RANDOMINIT;

  ParseParameters(argc, argv, 2, paraminfo);

  /* training set doesn't exist */
  if (!ExistFile(TSName))
    {
    ErrorMessage("\nERROR: File doesn't exist: %s\n\n", TSName);
    ExitProcessing(FATAL_ERROR);
    }

  /* result codebook exists and we are told not to overwrite */
  if (ExistFile(PAName) && !Value(OverWrite))
    {
    ErrorMessage("\nERROR: File already exists: %s\n\n", PAName);
    ExitProcessing(FATAL_ERROR);
    }

  ReadTrainingSet(TSName, &TS);

  /* codebook cannot contain more vectors than training set */
  if (BookSize(&TS) < Value(Clusters))
    {
    ErrorMessage("\nERROR: Number of vectors in training set ");
    ErrorMessage("(%d) < number of clusters ", BookSize(&TS));
    ErrorMessage("(%d)!\n\n", Value(Clusters));
    FreeCodebook(&TS);
    ExitProcessing(FATAL_ERROR);
    }

  CreateNewCodebook(&CB, Value(Clusters), &TS);
  CreateNewPartitioning(&P, &TS, Value(Clusters));
  CATCreateNewCodebook(&CCB, &CB, &TS);

  if (0)
    {
    ErrorMessage("ERROR: Allocating CATCODEBOOK failed!\n");
    FreeCodebook(&TS);
    FreeCodebook(&CB);
    FreePartitioning(&P);
    ExitProcessing(FATAL_ERROR);
    }

  initMethod = Value(InitMethod);

  PrintInitialData(Value(QuietLevel),
                   TSName,
                   PAName,
                   Value(Clusters),
                   Value(KMeansIterations),
                   Value(ClusteringMethod),
                   Value(InitMethod));


  if (Value(ClusteringMethod) == DTKMEDOIDS)
    {
    result = PerformKMedoidsDistanceTable(&P,
                                          &CCB,
                                          &TS,
                                          Value(KMeansIterations),
                                          Value(QuietLevel),
                                          Value(InitMethod));
    }
  else if (Value(ClusteringMethod) == KENTROPIES ||
           Value(ClusteringMethod) == DKENTROPIES ||
           Value(ClusteringMethod) == HKENTROPIES)
    {
    result = PerformKEntropies(&P,
                                &CCB,
                                &TS,
                                Value(KMeansIterations),
                                Value(QuietLevel),
                                Value(InitMethod),
                                Value(ClusteringMethod));
    }
  else if (Value(ClusteringMethod) == KHISTOGRAMS)
    {
    result = PerformKHistograms(&P,
                                &CCB,
                                &TS,
                                Value(KMeansIterations),
                                Value(QuietLevel),
                                Value(InitMethod),
                                Value(ClusteringMethod));
    }
  else
    {
    distanceMetric     = GetDistanceMetric    (Value(ClusteringMethod));
    representativeType = GetRepresentativeType(Value(ClusteringMethod));

    result = PerformCKM(&P,
                        &CCB,
                        &TS,
                        Value(KMeansIterations),
                        Value(QuietLevel),
                        distanceMetric,
                        representativeType,
                        initMethod);
    }

  if (result)
    {
    ErrorMessage("ERROR: Clustering failed!\n");
    FreeCodebook(&TS);
    FreeCodebook(&CB);
    FreePartitioning(&P);
    CATFreeCodebook(&CCB);
    ExitProcessing(FATAL_ERROR);
    }

  WritePartitioning(PAName, &P, &TS, Value(OverWrite));

  CATFreeCodebook(&CCB);

  FreeCodebook(&TS);
  FreeCodebook(&CB);
  FreePartitioning(&P);

  checkmemory();

  return 0;
  } /* main */

