/*
 $Revision: 1.13 $
 $Date: 2010/02/28 11:57:46 $
 $Author: apollan $
 $Name:  $
 $Id: ckm.c,v 1.13 2010/02/28 11:57:46 apollan Exp $
 */

/*-------------------------------------------------------------------*/
/* CKM.C        Antti Pollanen                                       */
/*                                                                   */
/* Special operations for categorical k-means algorithms             */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "ckm"
#define  VersionNumber  "Version 0.31"
#define  LastUpdated    "28.02.2009" /* AP */

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* AP   0.40  K-entropies speed-up and bug fix.                      */
/* AP   0.31  K-histograms iteration style fixed.                    */
/* AP   0.30  PerformKMedoidsDistanceTable implemented.              */
/*            Three variants of k-entropies algorithm.               */
/* AP   0.20  New iteration method implemented                       */
/* AP   0.11  KPRODUCTS moved to another branch                      */
/* AP   0.10  KPRODUCTS and KDISTRIBUTIONS added                     */
/* AP   0.01  Initial version                                        */
/*                                                                   */
/* ----------------------------------------------------------------- */

#include "cb.h"
#include "memctrl.h"
#include "math.h"
#include "interfc.h"
#include "random.h"

#include "ctk.h"
#include "ckm.h"

/* ========================== FUNCTIONS ============================ */

CATDISTANCETYPE GetDistanceMetric(CLUSTERINGMETHOD method)
  {
  CATDISTANCETYPE distanceMetric = 0;

  switch (method)
    {
    case KHISTOGRAMS:
      {
      distanceMetric = LINEARDISTANCE;
      break;
      }
    case KMEDOIDS:
    case STKMEDOIDS:
    case KMODES:
      {
      distanceMetric = SIMPLEMATCHINGDISTANCE;
      break;
      }
    case KDISTRIBUTIONS:
      {
      distanceMetric = JOINTPROBABILITYDISTANCE;
      break;
      }
    default:
      {
      ErrorMessage("ERROR: unsupported clustering method %d\n",
                   method);
      ExitProcessing(-1);
      }
    }

  return distanceMetric;
  } /* GetDistanceMetric */

/*-------------------------------------------------------------------*/

CATCENTROIDTYPE GetRepresentativeType(CLUSTERINGMETHOD method)
  {
  CATCENTROIDTYPE centroidType = 0;

  switch (method)
    {
    case KHISTOGRAMS:
    case KDISTRIBUTIONS:
      {
      centroidType = DISTRIBUTIONREPR;
      break;
      }
    case KMEDOIDS:
      {
      centroidType = MEDOIDREPR;
      break;
      }
    case KMODES:
      {
      centroidType = MODEREPR;
      break;
      }
    case STKMEDOIDS:
      {
      centroidType = STMEDOIDREPR;
      break;
      }
    default:
      {
      ErrorMessage("ERROR: unsupported clustering method %d\n",
                   method);
      ExitProcessing(-1);
      }
    }

  return centroidType;
  } /* GetRepresentativeType */

/*-------------------------------------------------------------------*/

double Entropy(int instances, int count)
  {
  double Probability;
  double Cost;

  if (instances == 0 || count == 0)
    {
    Cost = 0;
    }
  else
    {
    Probability = (double)instances / (double)count;
    Cost = -Probability * log(Probability);
    }

  return Cost;
  } /* Entropy */

/*-------------------------------------------------------------------*/

void KECountCosts(KEDATA* data, TRAININGSET* pTS, PARTITIONING* pP, int cluster)
  {
  int j, k, instances, count;

  CATCODEBOOK* pCCB = data->pCodebook;
  CATCODEBOOK* pDecreaseCost = data->pDecreaseCost;
  CATCODEBOOK* pIncreaseCost = data->pIncreaseCost;
  CODEBOOK* pCB = pCCB->pCodebook;
  double incC, decC;
  double oldDecC, oldIncC;
  double totalCost, totalDec, totalInc;

  count = Count(pCCB, cluster);

  for (j = 0; j < VectorSize(pCB); j++)
    {
    totalCost = 0; totalInc = 0; totalDec = 0;

    for (k = 0; k < Categories(pCCB, j); k++)
      {
      instances = CategoryInstances(pCCB, cluster, j, k);
      totalCost += Entropy(instances, count);
      totalInc  += Entropy(instances, count + 1);
      totalDec  += Entropy(instances, count - 1);
      }

    for (k = 0; k < Categories(pCCB, j); k++)
      {
      instances = CategoryInstances(pCCB, cluster, j, k);
      oldIncC = Entropy(instances, count + 1);
      oldDecC = Entropy(instances, count - 1);
      incC    = Entropy(instances + 1, count + 1);
      decC    = Entropy(instances - 1, count - 1);

      double inc = (count + 1) * (incC + totalInc - oldIncC) -
                    count      * (totalCost);
      double dec = (count - 1) * (decC + totalDec - oldDecC) -
                    count      * (totalCost);

      CategoryInstances(pIncreaseCost, cluster, j, k) = (int)(10000 * inc);
      CategoryInstances(pDecreaseCost, cluster, j, k) = (int)(10000 * dec);
      }
    }
  } /* KECountCosts */

/*-------------------------------------------------------------------*/

long KEDistance(TRAININGSET* pTS, KEDATA* data, int from, int to, int vector)
  {
  CATCODEBOOK *pCCB = data->pCodebook;
  CODEBOOK *pCB  = pCCB->pCodebook;
  CATCODEBOOK *pDecreaseCost = data->pDecreaseCost;
  CATCODEBOOK *pIncreaseCost = data->pIncreaseCost;
  int i, category;

  long distance = 0;
  for (i = 0; i < VectorSize(pCB); i++)
    {
    category = VectorScalar(pTS, vector, i);
    distance += CategoryInstances(pIncreaseCost, to,   i, category)
              + CategoryInstances(pDecreaseCost, from, i, category);
    }

  return distance;
  } /* KEDistance */

/*-------------------------------------------------------------------*/

void KEChangePartition(CATCODEBOOK *pCCB,
                        TRAININGSET *pTS,
                        PARTITIONING *pP,
                        int from,
                        int to,
                        int vector)
  {
  int i, value, count;

  if (to >= 0)
    {
    Count(pCCB, to)++;
    }

  if (from >= 0)
    {
    Count(pCCB, from)--;
    }

  for (i = 0; i < VectorSize(pTS); i++)
    {
    value = VectorScalar(pTS, vector, i);

    if (to >= 0)
      {
      count = CategoryInstances(pCCB, to, i, value)++;

      if (count == 0)
        {
        ClusterCategories(pCCB, to, i)++;
        }
      }

    if (from >= 0)
      {
      count = --CategoryInstances(pCCB, from, i, value);

      if (count == 0)
        {
        ClusterCategories(pCCB, from, i)--;
        }
      }
    }
  } /* KEChangePartition */

/*-------------------------------------------------------------------*/

double EntopyAfter(CATCODEBOOK *pCCB,
                   PARTITIONING *pP,
                   TRAININGSET *pTS,
                   int centroid,
                   int vector)
  {
  int old;
  double after;

  old = Map(pP, vector);

  if (old == centroid)
    {
    return 0.0;
    }

  KEChangePartition(pCCB, pTS, pP, old, centroid, vector);
  after  = CATClusterEntropy(pCCB, centroid) +
           CATClusterEntropy(pCCB, old);
  KEChangePartition(pCCB, pTS, pP, centroid, old, vector);

  return after;
  } /* EntopyAfter */

/*-------------------------------------------------------------------*/

int KEFindNearestVector(CATCODEBOOK* pCCB,
                        PARTITIONING *pP,
                        TRAININGSET *pTS,
                        int vector,
                        double* entropies)
  {
  int i, current;
  int MinIndex;
  double before, e, error = -EPSILON;
  CODEBOOK *pCB = pCCB->pCodebook;

  current = Map(pP, vector);
  MinIndex = current;

  if (Count(pCCB, current) > 1)
    {
    for (i = 0; i < BookSize(pCB); i++)
      {
      if (i != current)
        {
        before = entropies[current] + entropies[i];
        e = EntopyAfter(pCCB, pP, pTS, i, vector) - before;

        if (e < error)
          {
          error = e;
          MinIndex = i;
          }
        }
      }
    }

  return (MinIndex);
  } /* KEFindNearestVector */

/*-------------------------------------------------------------------*/

int QKEFindNearestVector(KEDATA* data,
                        PARTITIONING *pP,
                        TRAININGSET *pTS,
                        int vector)
  {
  int i, current;
  int MinIndex;
  double e, error = -EPSILON;
  CATCODEBOOK *pCCB = data->pCodebook;
  CODEBOOK *pCB = pCCB->pCodebook;

  current = Map(pP, vector);
  MinIndex = current;

  for (i = 0; i < BookSize(pCB); i++)
    {
    if (i != current)
      {
      e = KEDistance(pTS, data, current, i, vector);

      if (e < error)
        {
        error = e;
        MinIndex = i;
        }
      }
    }

  return (MinIndex);
  } /* QKEFindNearestVector */

/*-------------------------------------------------------------------*/

int IterateKEntropiesQuick(PARTITIONING *pP,
                           CATCODEBOOK *pCCB,
                           TRAININGSET *pTS,
                           int kmIter,
                           int quietLevel,
                           int *order,
                           int current,
                           int nochanges)
  {
  int i, nearest, vector;
  int size = BookSize(pTS);
  CODEBOOK* pCB = pCCB->pCodebook;
  int fullIter = 0;

  KEDATA data;
  CATCODEBOOK Increase, Decrease;
  data.pIncreaseCost = &(Increase);
  data.pDecreaseCost = &(Decrease);
  data.pCodebook = pCCB;
  CATCreateNewCodebook(data.pDecreaseCost, pCB, pTS);
  CATCreateNewCodebook(data.pIncreaseCost, pCB, pTS);

  for (i = 0; i < BookSize(pCB); i++)
    {
    KECountCosts(&data, pTS, pP, i);
    }

  for (i = current; (fullIter < kmIter || kmIter == 0) && nochanges < size; i++)
    {
    CATPrintV(quietLevel, DEBUGMESSAGE, "Iteration %d - %d\n", fullIter, i+1);
    vector = order[i % size];

    nearest = QKEFindNearestVector(&data, pP, pTS, vector);
    if (nearest == Map(pP, vector))
      {
      nochanges++;
      }
    else
      {
      CATPrintV(quietLevel, DEBUGMESSAGE,
                "Vector %d moved from cluster %d to cluster %d\n",
                vector, Map(pP, vector), nearest);

      int previous = Map(pP, vector);
      CATChangePartition(pCCB, pTS, pP, Map(pP, vector), nearest, vector);
      KECountCosts(&data, pTS, pP, previous);
      KECountCosts(&data, pTS, pP, nearest);
      nochanges = 0;
      }

    if (i == size - 1)
      {
      i = -1;
      fullIter++;
      }
    }

  CATFreeCodebook(&Decrease);
  CATFreeCodebook(&Increase);


  PrintMessage("Codebook size %d, Iterations %d - %d\n",
               BookSize(pCCB->pCodebook), fullIter, i);

  return nochanges;
  } /* IterateKEntropiesQuick */

/*-------------------------------------------------------------------*/

int IterateKEntropies(PARTITIONING *pP,
                      CATCODEBOOK *pCCB,
                      TRAININGSET *pTS,
                      int kmIter,
                      int quietLevel,
                      int *order,
                      int current,
                      int nochanges)
  {
  int i, nearest, vector;
  double* entropies;
  int size = BookSize(pTS);
  int fullIter = 0;
  CODEBOOK* pCB = pCCB->pCodebook;

  entropies = (double*)allocate(BookSize(pCB)*sizeof(double));

  for (i = 0; i < BookSize(pCB); i++)
    {
    entropies[i] = CATClusterEntropy(pCCB, i);
    }

  CATPrint(quietLevel, INFOMESSAGE, "InitialSolution created\n");
  nochanges = 0;

  for (i = current; (fullIter < kmIter || kmIter == 0) && nochanges < size; i++)
    {
    CATPrintV(quietLevel, DEBUGMESSAGE, "Iteration %d - %d\n", fullIter, i+1);
    vector = order[i % size];

    int previous = Map(pP, vector);
    nearest = KEFindNearestVector(pCCB, pP, pTS, vector, entropies);

    if (nearest == previous)
      {
      nochanges++;
      }
    else
      {
      CATPrintV(quietLevel, DEBUGMESSAGE,
                "Vector %d moved from cluster %d to cluster %d\n",
                vector, Map(pP, vector), nearest);
      CATChangePartition(pCCB, pTS, pP, previous, nearest, vector);
      entropies[previous] = CATClusterEntropy(pCCB, previous);
      entropies[nearest]  = CATClusterEntropy(pCCB, nearest);
      nochanges = 0;
      }

    if (i == size - 1)
      {
      i = -1;
      fullIter++;
      }
    }

  PrintMessage("Codebook size %d, Iterations %d - %d\n",
               BookSize(pCCB->pCodebook), fullIter, i);

  deallocate(entropies);
  return nochanges;
  } /* IterateKEntropies */

/*-------------------------------------------------------------------*/

int PerformKEntropies   (PARTITIONING *pP,
                         CATCODEBOOK *pCCB,
                         TRAININGSET *pTS,
                         int kmIter,
                         int quietLevel,
                         CATINITMETHOD initMethod,
                         CLUSTERINGMETHOD method)
  {
  CATPrint(quietLevel, INFOMESSAGE, "Start K-Entropies\n");
  int* order;
  int nochanges, i;
  int size = BookSize(pTS);

  order = (int*)allocate(size*sizeof(int));
  for (i = 0; i < size; i++)
    {
    order[i] = i;
    }

  if (initMethod != SIMPLEINIT)
    {
    ShuffleMemory((void*)order, size, sizeof(int));
    }

  CATClearDivisions(pCCB);
  CATGenerateInitialSolution(pP, pCCB, pTS, initMethod);
  CATGenerateOptimalCodebook(pP, pTS, pCCB, DISTRIBUTIONREPR);

  switch (method)
    {
    case KENTROPIES:
      IterateKEntropies(pP, pCCB, pTS, kmIter, quietLevel, order, 0, 0);
      break;
    case DKENTROPIES:
      IterateKEntropiesQuick(pP, pCCB, pTS, kmIter, quietLevel, order, 0, 0);
      break;
    case HKENTROPIES:
      nochanges = IterateKEntropies(pP, pCCB, pTS, 1, quietLevel, order, 0, 0);

      if (kmIter != 1)
        {
        if (kmIter > 1)
          {
          kmIter--;
          }

        IterateKEntropiesQuick(pP, pCCB, pTS, kmIter, quietLevel, order, 0, nochanges);
        }
      break;
    default:
      ErrorMessage("Invalid clustering method %d!\n", method);
      return 1;
      break;
    }

  CATPrint(quietLevel, INFOMESSAGE, "End K-Entropies\n");
  return 0;
  } /* PerformKEntropies */

/*-------------------------------------------------------------------*/

int PerformKMedoidsDistanceTable(PARTITIONING *pP,
                                 CATCODEBOOK *pCCB,
                                 TRAININGSET *pTS,
                                 int kmIter,
                                 int quietLevel,
                                 CATINITMETHOD initMethod)
  {
  CATPrint(quietLevel, INFOMESSAGE, "Distance table k-medoids started\n");

  int i, j, changes;

  UTM utm;
  UTMCreate(&utm, BookSize(pTS));
  int *distances = (int*)allocate(utm.Size * sizeof(int));
  int Index = 0;

  for (i = 0; i < BookSize(pTS); i++)
    {
    for (j = i + 1; j < BookSize(pTS); j++)
      {
      distances[Index] = -1;
      UTMSet(&utm, i, j, &(distances[Index]));
      Index++;
      }
    }

  CATClearDivisions(pCCB);
  CATGenerateInitialSolution(pP, pCCB, pTS, initMethod);
  CATGenerateMedoids(pP, pTS, pCCB, &utm);

  CATPrint(quietLevel, INFOMESSAGE, "InitialSolution created\n");

  for (i = 0; i < kmIter || kmIter == 0; i++)
    {
    changes = CATGenerateOptimalPartition(pCCB, pTS, pP, SIMPLEMATCHINGDISTANCE);
    CATPrintIterationInfo(quietLevel, i + 1, changes);

    if (changes == 0)
      {
      break;
      }

    CATGenerateMedoids(pP, pTS, pCCB, &utm);

    for (j = 0; j < BookSize(pCCB->pCodebook); j++)
      {
      CATPrintV(quietLevel, DEBUGMESSAGE, "Prototype %d ( %d ):\n",
                j, Count(pCCB, j));
      }
    }

  deallocate(distances);
  CATPrint(quietLevel, INFOMESSAGE, "Distance table k-medoids completed\n");

  PrintMessage("Codebook size %d, Iterations %d\n",
               BookSize(pCCB->pCodebook), i);

  return 0;
  } /* PerformKMedoidsDistanceTable */

/*-------------------------------------------------------------------*/


int PerformKHistograms  (PARTITIONING *pP,
                         CATCODEBOOK *pCCB,
                         TRAININGSET *pTS,
                         int kmIter,
                         int quietLevel,
                         CATINITMETHOD initMethod,
                         CLUSTERINGMETHOD method)
  {
  int i, nearest, vector;
  int fullIter = 0;
  int* order;
  int nochanges = 0;
  int size = BookSize(pTS);

  order = (int*)allocate(size*sizeof(int));
  for (i = 0; i < size; i++)
    {
    order[i] = i;
    }

  if (initMethod != SIMPLEINIT)
    {
    ShuffleMemory((void*)order, size, sizeof(int));
    }

  CATClearDivisions(pCCB);
  CATGenerateInitialSolution(pP, pCCB, pTS, initMethod);
  CATGenerateOptimalCodebook(pP, pTS, pCCB, DISTRIBUTIONREPR);

  for (i = 0; (fullIter < kmIter || kmIter == 0) && nochanges < size; i++)
    {
    CATPrintV(quietLevel, DEBUGMESSAGE, "Iteration %d - %d\n", fullIter, i+1);
    vector = order[i % size];

    int previous = Map(pP, vector);
    nearest = CATFindNearestVector(pCCB, pP, pTS, vector, LINEARDISTANCE);

    if (nearest == previous)
      {
      nochanges++;
      }
    else
      {
      CATPrintV(quietLevel, DEBUGMESSAGE,
                "Vector %d moved from cluster %d to cluster %d\n",
                vector, Map(pP, vector), nearest);
      CATChangePartition(pCCB, pTS, pP, previous, nearest, vector);
      nochanges = 0;
      }

    if (i == size - 1)
      {
      i = -1;
      fullIter++;
      }
    }

  PrintMessage("Codebook size %d, Iterations %d - %d\n",
               BookSize(pCCB->pCodebook), fullIter, i);

  deallocate(order);
  return 0;
  } /* PerformKHistograms */

/*-------------------------------------------------------------------*/

