/*
$Revision: 1.9 $
$Date: 2009/10/01 12:34:04 $
$Author: apollan $
$Name:  $
$Id: cbctk.c,v 1.9 2009/10/01 12:34:04 apollan Exp $
*/

/*-------------------------------------------------------------------*/
/* CBCTK.C     Antti Pollanen                                        */
/*                                                                   */
/* Console interface for Categorial data ToolKit                     */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "cbctk"
#define  VersionNumber  "Version 0.31"
#define  LastUpdated    "01.10.2009" /* AP */
#define  MAXFILENAME    64
#define  FACTFILE       "cbctk.fac"

/*-------------------------------------------------------------------*/

#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"

#include "ctk.h"

#include "math.h"
#include "parametr.c"


/* ========================== PROTOTYPES =========================== */

void PrintInitialData(int quietLevel, char *TSName, char *PAName,
  int clus, int kmIter, int rlsIter);

void PrintInfo(void);

int main(int argc, char* argv[]);


/* ========================== FUNCTIONS ============================ */

void PrintInitialData(int quietLevel, char *TSName, char *PAName,
  int clus, int kmIter, int rlsIter)
  {
  if (quietLevel > 1)
    {
    PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
    PrintMessage("Training Set\t\t = %s\n", TSName);
    PrintMessage("Result Partitioning\t = %s\n", PAName);
    PrintMessage("Codebook size\t\t = %d\n", clus);
    PrintMessage("K-means iterations\t = %d\n", kmIter);
    PrintMessage("RLS iterations\t\t = %d\n", rlsIter);
    PrintMessage("\n");
    }
  }  /* PrintInitialData() */

/*-------------------------------------------------------------------*/

void PrintInfo(void)
  {
  PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
  PrintMessage("Console interface for the categorical toolkit.\n");
  PrintMessage("Usage: %s [%coption] ", ProgName, OPTION_SYMBOL);
  PrintMessage("<training set> <result partitioning>\n");
  PrintMessage("For example: %s bridge tmp\n", ProgName);
  PrintMessage("\n  Options:\n");
  PrintOptions();
  PrintMessage("\n");
  }  /* PrintInfo() */

/*-------------------------------------------------------------------*/

int main(int argc, char* argv[])
  {
  TRAININGSET   TS;
  CODEBOOK      CB;
  CATCODEBOOK   CCB;
  PARTITIONING  P;
  char          TSName[MAXFILENAME] = {'\0'};
  char          PAName[MAXFILENAME] = {'\0'};
  ParameterInfo paraminfo[2] = { { TSName,  FormatNameTS, 0, INFILE },
                                 { PAName,  FormatNamePA, 0, OUTFILE } };
  int result;

  ParseParameters(argc, argv, 2, paraminfo);

  /* training set doesn't exist */
  if (!ExistFile(TSName))
    {
    ErrorMessage("\nERROR: File doesn't exist: %s\n\n", TSName);
    ExitProcessing(FATAL_ERROR);
    }

  /* result codebook exists and we are told not to overwrite */
  if (ExistFile(PAName) && !Value(OverWrite))
    {
    ErrorMessage("\nERROR: File already exists: %s\n\n", PAName);
    ExitProcessing(FATAL_ERROR);
    }

  ReadTrainingSet(TSName, &TS);

  /* codebook cannot contain more vectors than training set */
  if (BookSize(&TS) < Value(Clusters))
    {
    ErrorMessage("\nERROR: Number of vectors in training set ");
    ErrorMessage("(%d) < number of clusters ", BookSize(&TS));
    ErrorMessage("(%d)!\n\n", Value(Clusters));
    FreeCodebook(&TS);
    ExitProcessing(FATAL_ERROR);
    }

  CreateNewCodebook(&CB, Value(Clusters), &TS);
  CreateNewPartitioning(&P, &TS, Value(Clusters));
  CATCreateNewCodebook(&CCB, &CB, &TS);

  if (0)
    {
    ErrorMessage("ERROR: Allocating CATCODEBOOK failed!\n");
    FreeCodebook(&TS);
    FreeCodebook(&CB);
    FreePartitioning(&P);
    ExitProcessing(FATAL_ERROR);
    }

  PrintInitialData(Value(QuietLevel),
                   TSName,
                   PAName,
                   Value(Clusters),
                   Value(KMeansIterations),
                   Value(RLSIterations));

  if (Value(RLSIterations) == 1)
    {
    result = PerformCKM(&P,
                       &CCB,
                       &TS,
                       Value(KMeansIterations),
                       Value(QuietLevel),
                       Value(DistanceMetric),
                       Value(RepresentativeType));
    }
  else
    {
    result = PerformRLS(&P,
                        &CCB,
                        &TS,
                        Value(KMeansIterations),
                        Value(RLSIterations),
                        Value(QuietLevel),
                        Value(DistanceMetric));
    }

  if (result)
    {
    ErrorMessage("ERROR: Clustering failed!\n");
    FreeCodebook(&TS);
    FreeCodebook(&CB);
    FreePartitioning(&P);
    CATFreeCodebook(&CCB);
    ExitProcessing(FATAL_ERROR);
    }

  // WriteCodebook(CBName, &CB, Value(OverWrite));
  WritePartitioning(PAName, &P, &TS, Value(OverWrite));

  CATFreeCodebook(&CCB);

  FreeCodebook(&TS);
  FreeCodebook(&CB);
  FreePartitioning(&P);

  checkmemory();

  return 0;
  } /* main */

