/*
 $Revision: 1.26 $
 $Date: 2010/02/08 18:49:11 $
 $Author: apollan $
 $Name:  $
 $Id: ctk.c,v 1.26 2010/02/08 18:49:11 apollan Exp $
 */

/*-------------------------------------------------------------------*/
/* CTK.C        Antti Pollanen                                       */
/*                                                                   */
/* Categorial data ToolKit                                           */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName        "CTK"
#define VersionNumber   "Version 0.61"
#define LastUpdated     "08.02.2010" /* AP */

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* AP   0.61  Constant time ClusterCategories implemented            */
/* AP   0.60  CATChangePartition extra condition added               */
/* AP   0.59  InitRandom method added                                */
/* AP   0.58  QMEDOIDSREPR renamed to STMEDOIDSREPR                  */
/* AP   0.57  CATMergeClusters bug fixed.                            */
/* AP   0.56  K-medoids improved, quick k-medoids implemented        */
/* AP   0.55  Tie-break issues solved                                */
/* AP   0.54  InitMethod selection to PerformCKM. PerformRLS removed */
/* AP   0.53  Init methods improved                                  */
/* AP   0.52  Testing distances moved to another branch              */
/* AP   0.51  CATExpectedEntropy added                               */
/* AP   0.50  Critical bugs in distance calculation and random       */
/*            initialization  fixed                                  */
/* AP   0.43  JOINTPROBABILITYDISTANCE added                         */
/* AP   0.42  PRODUCTDISTANCE added                                  */
/* AP   0.41  CATDatasetEntropy function added                       */
/* AP   0.40  CBCKM separated and tools added (map, sample, scale)   */
/* AP   0.32  CATMergeClusters-bug fixed (check if cluster2 is last) */
/* AP   0.31  Stardardization for coding and commenting              */
/* AP   0.30  Representative and distance selection by switch        */
/* AP   0.21  Printing routines for cluster representatives and      */
/*            codebook                                               */
/* AP   0.20  Interface changed, PrintSimpleMessage-method added     */
/* AP   0.12  Entropy distance and RLS iterations added              */
/* AP   0.11  Minor changed for categorical algorithms               */
/* AP   0.10  CATCODEBOOK Structure modified                         */
/* AP   0.01  Initial version                                        */
/*                                                                   */
/* ----------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>
#include "parametr.h"
#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "random.h"

#include "ctk.h"
#include "interfc.h"

#include "math.h"


/* ========================== PROTOTYPES =========================== */

double CATVectorDistanceLinear     (CATCODEBOOK *pCCB,
                                    PARTITIONING *pP,
                                    TRAININGSET *pTS,
                                    int centroid,
                                    int vector);

void CATSelectSimplePartitions     (TRAININGSET *pTS,
                                    CATCODEBOOK *pCCB,
                                    PARTITIONING *pP);

void CATSelectRandomPartitions     (TRAININGSET *pTS,
                                    CATCODEBOOK *pCCB,
                                    PARTITIONING *pP);

void CATSelectRandomRepresentatives(TRAININGSET *pTS,
                                    CATCODEBOOK *pCCB,
                                    PARTITIONING *pP);

void PrintInitialData              (int quietLevel,
                                    char *TSName,
                                    char *CBName,
                                    int clus,
                                    int kmIter);


/* ========================== FUNCTIONS ============================ */

void CATScaleTrainingSet(TRAININGSET* in, TRAININGSET* out)
  {
  int a, v, i;
  int max, min;
  int* values;

  for (a = 0; a < VectorSize(in); a++)
    {
    min = max = VectorScalar(in, 0, a);

    // Find out maximum and minimum for value-array allocation.
    for (v = 0; v < BookSize(in); v++)
      {
      if (VectorScalar(in, v, a) < min)
        {
        min = VectorScalar(in, v, a);
        }

      if (VectorScalar(in, v, a) > max)
        {
        max = VectorScalar(in, v, a);
        }
      }

    values = (int*) allocate((max - min + 2) * sizeof(int));

    // Find out number of value instances.
    for (i = min; i <= max; i++)
      {
      values[i - min] = 0;
      }

    for (v = 0; v < BookSize(in); v++)
      {
      values[VectorScalar(in, v, a) - min] = 1;
      }

    // Find out new values.
    v = 1;
    for (i = min; i <= max; i++)
      {
      if (values[i - min] != 0)
        {
        values[i - min] = v++;
        }
      }

    // Set new values to new training set.
    for (v = 0; v < BookSize(in); v++)
      {
      i = values[VectorScalar(in, v, a) - min];
      VectorScalar(out, v, a) = i - 1;
      }

    deallocate(values);
    values = 0;
    }
  } /* CATScaleTrainingSet */

/*-------------------------------------------------------------------*/

int CATCreateNewCodebook(CATCODEBOOK *pCCB,
                         CODEBOOK *pCB,
                         TRAININGSET *pTS)
  {
  int cats, i, j, k;
  CENTROIDINFO *info;
  DIVISION *pDivision;

  pCCB->pCodebook = pCB;
  pCCB->CentroidInformation = NULL;
  pCCB->Categories = NULL;

  /* Calculates number of categories. */
  pCCB->Categories = (int*) allocate(VectorSize(pCB) * sizeof(int));

  if (pCCB->Categories == NULL)
    {
    CATFreeCodebook(pCCB);
    return 1;
    }

  for (i = 0; i < VectorSize(pCB); i++)
    {
    cats = 0;

    for (j = 0; j < BookSize(pTS); j++)
      {
      if (VectorScalar(pTS, j, i) > cats)
        {
        cats = VectorScalar(pTS, j, i);
        }
      }

    pCCB->Categories[i] = cats + 1;
    }

  /* Allocates needed memory. */
  pCCB->CentroidInformation = (CENTROIDINFO*) allocate(BookSize(pCB)
      * sizeof(CENTROIDINFO));

  if (pCCB->CentroidInformation == NULL)
    {
    CATFreeCodebook(pCCB);
    return 1;
    }

  for (i = 0; i < BookSize(pCB); i++)
    {
    info = &(pCCB->CentroidInformation[i]);
    info->Divisions = (DIVISION*) allocate(VectorSize(pCB) * sizeof(DIVISION));
    info->Categories = (int*) allocate(VectorSize(pCB) * sizeof(int));

    if (info->Divisions == NULL || info->Categories == NULL)
      {
      CATFreeCodebook(pCCB);
      return 1;
      }

    for (j = 0; j < VectorSize(pCB); j++)
      {
      pDivision = &(info->Divisions[j]);
      pDivision->Instances = (int*) allocate(Categories(pCCB, j)
          * sizeof(DIVISION));

      if (pDivision->Instances == NULL)
        {
        CATFreeCodebook(pCCB);
        return 1;
        }

      for (k = 0; k < Categories(pCCB, j); k++)
        {
        pDivision->Instances[k] = 0;
        // CategoryInstances(MCB, i, j, k) = 0;
        }
      }
    }

  return 0; // everything is ok.
  } /* CreateNewCATCodebook */

/*-------------------------------------------------------------------*/

void CATFreeCodebook(CATCODEBOOK* pCCB)
  {
  int i, j;
  CODEBOOK *pCB;
  CENTROIDINFO *info;
  DIVISION *divisions;

  if (pCCB != NULL)
    {
    pCB = pCCB->pCodebook;

    if (pCCB->CentroidInformation != NULL)
      {
      for (i = 0; i < BookSize(pCB); i++)
        {
        info = &(pCCB->CentroidInformation[i]);

        if (info != NULL)
          {
          divisions = info->Divisions;

          for (j = 0; j < VectorSize(pCB); j++)
            {
            if (divisions[j].Instances != NULL)
              {
              deallocate(divisions[j].Instances);
              }
            }

          deallocate(divisions);
          deallocate(info->Categories);
          }
        }

      deallocate(pCCB->CentroidInformation);
      }

    if (pCCB->Categories != NULL)
      {
      deallocate(pCCB->Categories);
      }
    }
  } /* CATFreeCodebook */

/*-------------------------------------------------------------------*/

void CATClearDivisions(CATCODEBOOK *pCCB)
  {
  int i, j, k;
  CODEBOOK *pCB = pCCB->pCodebook;

  for (i = 0; i < BookSize(pCB); i++)
    {
    Count(pCCB, i) = 0;

    for (j = 0; j < VectorSize(pCB); j++)
      {
      ClusterCategories(pCCB, i, j) = 0;

      for (k = 0; k < Categories(pCCB, j); k++)
        {
        CategoryInstances(pCCB, i, j, k) = 0;
        }
      }
    }
  } /* CATClearDivisions */

/*-------------------------------------------------------------------*/

void CATChangePartition(CATCODEBOOK *pCCB,
                        TRAININGSET *pTS,
                        PARTITIONING *pP,
                        int from,
                        int to,
                        int vector)
  {
  // Refuse to change last element.
  if (from != -1 && Count(pCCB, from) == 1)
    {
    return;
    }

  int i, value, count;

  if (to >= 0)
    {
    Count(pCCB, to)++;
    }

  if (from >= 0)
    {
    Count(pCCB, from)--;
    }

  for (i = 0; i < VectorSize(pTS); i++)
    {
    value = VectorScalar(pTS, vector, i);

    if (to >= 0)
      {
      count = CategoryInstances(pCCB, to, i, value)++;

      if (count == 0)
        {
        ClusterCategories(pCCB, to, i)++;
        }
      }

    if (from >= 0)
      {
      count = --CategoryInstances(pCCB, from, i, value);

      if (count == 0)
        {
        ClusterCategories(pCCB, from, i)--;
        }
      }
    }

  if (to >= 0 && Map(pP, vector) != to)
    {
    ChangePartition(pTS, pP, to, vector);
    }
  } /* CATChangePartition */

/*-------------------------------------------------------------------*/

void CATMergeClusters(CATCODEBOOK *pCCB,
                      TRAININGSET *pTS,
                      PARTITIONING *pP,
                      int cluster1,
                      int cluster2)
  {
  int a, k;
  int last = pP->PartitionCount - 1;
  int swap;

  if (cluster1 != cluster2)
    {
    if (cluster1 > cluster2)
      {
      swap = cluster1;
      cluster1 = cluster2;
      cluster2 = swap;
      }

    Count(pCCB, cluster1) += Count(pCCB, cluster2);
    Count(pCCB, cluster2) = Count(pCCB, last);

    for (a = 0; a < VectorSize(pTS); a++)
      {
      for (k = 0; k < Categories(pCCB, a); k++)
        {
        CategoryInstances(pCCB, cluster1, a, k) +=
          CategoryInstances(pCCB,cluster2, a, k);

        if (cluster2 != last)
          {
          CategoryInstances(pCCB, cluster2, a, k) =
            CategoryInstances(pCCB, last, a, k);

          CategoryInstances(pCCB, last, a, k) = 0;
          Count(pCCB, last) = 0;
          }
        }
      }

    JoinPartitions(pTS, pP, cluster1, cluster2);
    }
  } /* CATMergeClusters */

/*-------------------------------------------------------------------*/

double CATVectorDistanceSimpleMatching(CATCODEBOOK *pCCB, PARTITIONING *pP,
    TRAININGSET *pTS, int centroid, int vector)
  {
  int i;
  int category1, category2;
  int different = 0;
  CODEBOOK* pCB = pCCB->pCodebook;

  for (i = 0; i < VectorSize(pCCB->pCodebook); i++)
    {
    category1 = VectorScalar(pCB, centroid, i);
    category2 = VectorScalar(pTS, vector, i);

    if (category1 != category2)
      {
      different++;
      }
    }

  return different;
  } /* CATVectorDistanceSimpleMatching */

/*-------------------------------------------------------------------*/

double CATVectorDistanceJointProbability(CATCODEBOOK *pCCB,
                                         PARTITIONING *pP,
                                         TRAININGSET *pTS,
                                         int centroid,
                                         int vector)
  {
  int i;
  int category;
  double product = 0;
  double p;

  for (i = 0; i < VectorSize(pCCB->pCodebook); i++)
    {
    category = VectorScalar(pTS, vector, i);

    p = (double) (CategoryInstances(pCCB, centroid, i, category)
                  + 1.0 / Categories(pCCB, i))
      / (double) (Count(pCCB, centroid) + 1);
    product += log(p);
    }

  return -product;
  } /* CATVectorDistanceJointProbability */

/*-------------------------------------------------------------------*/

double CATVectorDistanceLinear(CATCODEBOOK *pCCB,
                               PARTITIONING *pP,
                               TRAININGSET *pTS,
                               int centroid,
                               int vector)
  {
  int i;
  int category;
  double sum = 0;
  double p;

  for (i = 0; i < VectorSize(pCCB->pCodebook); i++)
    {
    category = VectorScalar(pTS, vector, i);
    sum += 1;

    if (Count(pCCB, centroid) != 0)
      {
      p = (double) CategoryInstances(pCCB, centroid, i, category)
        / Count(pCCB, centroid);
      sum -= p;
      }
    }

  return sum;
  } /* CATVectorDistanceLinear */

/*-------------------------------------------------------------------*/

double CATVectorDistanceEntropy(CATCODEBOOK *pCCB, PARTITIONING *pP,
    TRAININGSET *pTS, int centroid, int vector)
  {
  int old;
  double before, after;

  old = Map(pP, vector);

  if (old == centroid)
    {
    return 0.0;
    }

  before = CATClusterEntropy(pCCB, centroid) +
           CATClusterEntropy(pCCB, old);
  CATChangePartition(pCCB, pTS, pP, old, centroid, vector);
  after  = CATClusterEntropy(pCCB, centroid) +
           CATClusterEntropy(pCCB, old);
  CATChangePartition(pCCB, pTS, pP, centroid, old, vector);

  return after - before;
  } /* CATVectorDistanceEntropy */

/*-------------------------------------------------------------------*/

double CATVectorDistance(CATCODEBOOK *pCCB,
                         PARTITIONING *pP,
                         TRAININGSET *pTS,
                         int centroid,
                         int vector,
                         CATDISTANCETYPE disttype)
  {

  switch (disttype)
    {
  case SIMPLEMATCHINGDISTANCE:
    {
    return CATVectorDistanceSimpleMatching(pCCB, pP, pTS, centroid, vector);
    }
  case JOINTPROBABILITYDISTANCE:
    {
    return CATVectorDistanceJointProbability(pCCB, pP, pTS, centroid, vector);
    }
  case LINEARDISTANCE:
    {
    return CATVectorDistanceLinear(pCCB, pP, pTS, centroid, vector);
    }
  case ENTROPYDISTANCE:
    {
    return CATVectorDistanceEntropy(pCCB, pP, pTS, centroid, vector);
    }
    }

  ErrorMessage("ERROR: unsupported vector distance %d\n", disttype);
  ExitProcessing(-1);

  return 0; /* never reached */
  } /* CATVectorDistance */

/*-------------------------------------------------------------------*/

int CATFindNearestVector(CATCODEBOOK *pCCB,
                         PARTITIONING *pP,
                         TRAININGSET *pTS,
                         int vector,
                         CATDISTANCETYPE disttype)
  {
  int i;
  int MinIndex = 0;
  double e, error = 0;
  CODEBOOK *pCB = pCCB->pCodebook;

  for (i = 0; i < BookSize(pCB); i++)
    {
    e = CATVectorDistance(pCCB, pP, pTS, i, vector, disttype);

    // This gives advantage to current cluster in ties
    if (i == Map(pP, vector))
      {
      e = e - 0.00001;
      }

    if (i == 0 || e < error)
      {
      error = e;
      MinIndex = i;
      }
    }

  return (MinIndex);
  } /* CATFindNearestVector */

/*-------------------------------------------------------------------*/

void CATGenerateInitialSolution(PARTITIONING *pP,
                                CATCODEBOOK *pCCB,
                                TRAININGSET *pTS,
                                CATINITMETHOD initMethod)
  {
  switch (initMethod)
    {
  case SIMPLEINIT:
    CATSelectSimplePartitions(pTS, pCCB, pP);
    break;
  case RANDOMINIT:
    CATSelectRandomPartitions(pTS, pCCB, pP);
    break;
    }
  } /* CATGenerateInitialSolution */

/*-------------------------------------------------------------------*/

void CATSelectSimplePartitions(TRAININGSET *pTS,
                               CATCODEBOOK *pCCB,
                               PARTITIONING *pP)
  {
  int i, j, seq, extra;
  int bookSize = BookSize(pCCB->pCodebook);

  seq = BookSize(pTS) / bookSize;
  extra = BookSize(pTS) - bookSize * seq;

  for (i = 0; i < bookSize; i++)
    {
    if (i < extra)
      {
      for (j = i * (seq + 1); j < (i + 1) * (seq + 1); j++)
        {
        CATChangePartition(pCCB, pTS, pP, -1, i, j);
        }
      }
    else
      {
      for (j = i * seq + extra; j < (i + 1) * seq + extra; j++)
        {
        CATChangePartition(pCCB, pTS, pP, -1, i, j);
        }
      }
    }
  } /* CATSelectSimplePartitions */


/*-------------------------------------------------------------------*/

void CATSelectRandomPartitions(TRAININGSET *pTS,
                               CATCODEBOOK *pCCB,
                               PARTITIONING *pP)
  {
  InitRandom();
  int i, cluster;
  int* order = (int*)allocate(BookSize(pTS) * sizeof(int));

  for (i = 0; i < BookSize(pTS); i++)
    {
    order[i] = i;
    }

  ShuffleMemory( (void*)order, BookSize(pTS), sizeof(int) );

  cluster = 0;
  for (i = 0; i < BookSize(pTS); i++)
    {
    CATChangePartition(pCCB, pTS, pP, -1, cluster, order[i]);
    cluster++;

    if (cluster >= BookSize(pCCB->pCodebook))
      {
      cluster = 0;
      }
    }

  deallocate(order);
  } /* CATSelectRandomPartitions */

/*-------------------------------------------------------------------*/

void CATGenerateDistributions(PARTITIONING *pP,
                              TRAININGSET *pTS,
                              CATCODEBOOK *pCCB)
  {
  int i;
  int prototype;

  CATClearDivisions(pCCB);

  for (i = 0; i < BookSize(pTS); i++)
    {
    prototype = Map(pP, i);
    CATChangePartition(pCCB, pTS, pP, -1, prototype, i);
    }
  } /* CATGenerateDistributions */

/*-------------------------------------------------------------------*/

int CATMedoidDistance(CATCODEBOOK *pCCB,
                      PARTITIONING *pP,
                      TRAININGSET *pTS,
                      int v1,
                      int v2)
    {
    int i;
    int category1, category2;
    int different = 0;

    for (i = 0; i < VectorSize(pCCB->pCodebook); i++)
      {
      category1 = VectorScalar(pTS, v1, i);
      category2 = VectorScalar(pTS, v2, i);

      if (category1 != category2)
        {
        different++;
        }
      }

    return different;
    } /* CATMedoidDistance */

/*-------------------------------------------------------------------*/

void FindMedoid(PARTITIONING *pP,
                TRAININGSET *pTS,
                CATCODEBOOK *pCCB,
                int prototype,
                UTM* distances)
  {
  int i, j;
  double sum, minSum;
  int distance, minIndex;
  int previousMedoid;
  CODEBOOK *pCB = pCCB->pCodebook;

  minIndex = -1;
  minSum = -1;

  i = FirstVector(pP, prototype);

  while( !EndOfPartition(i) )
    {
    j = FirstVector(pP, prototype);
    sum = 0;

    while( !EndOfPartition(j) )
      {
      if (i != j)
        {
        if (distances == 0)
          {
          distance = CATMedoidDistance(pCCB, pP, pTS, i, j);
          }
        else
          {
          int* pDistance = (int*)UTMGet(distances, i, j);

          if (*pDistance < 0)
            {
            *pDistance = CATMedoidDistance(pCCB, pP, pTS, i, j);
            }

          distance = *pDistance;
          }

        sum += distance;
        }

      j = NextVector(pP, j);
      }

    previousMedoid = !CompareVectors(Vector(pCB,prototype), Vector(pTS, i), VectorSize(pCB));

    if (sum < minSum || (sum == minSum && previousMedoid)|| minIndex == -1)
      {
      minSum = sum;
      minIndex = i;
      }

    i = NextVector(pP, i);
    }

  CopyVector(Vector(pTS, minIndex), Vector(pCB,prototype), VectorSize(pCB));
  } /* FindMedoid */

/*-------------------------------------------------------------------*/

void CATGenerateMedoids(PARTITIONING *pP,
                        TRAININGSET *pTS,
                        CATCODEBOOK *pCCB,
                        UTM* distances)
  {
  int c;
  int size = BookSize(pCCB->pCodebook);

  for (c = 0; c < size; c++)
    {
    FindMedoid(pP, pTS, pCCB, c, distances);
    }
  } /* CATGenerateMedoids */

/*-------------------------------------------------------------------*/

void FindMedoidQuick(PARTITIONING *pP,
                             TRAININGSET *pTS,
                             CATCODEBOOK *pCCB,
                             int prototype)
  {
  int i, j;
  int match;
  int max, maxIndex;
  int category, previousMedoid;
  CODEBOOK *pCB = pCCB->pCodebook;

  max = -1;
  maxIndex = -1;

  i = FirstVector(pP, prototype);
  while( !EndOfPartition(i) )
    {
    match = 0;

    for (j = 0; j < VectorSize(pCB); j++)
      {
      category = VectorScalar(pTS, i, j);
      match += CategoryInstances(pCCB, prototype, j, category);
      }

    previousMedoid = !CompareVectors(Vector(pCB,prototype), Vector(pTS, i), VectorSize(pCB));

    if (match > max || (match == max && previousMedoid))
      {
      max = match;
      maxIndex = i;
      }

    i = NextVector(pP, i);
    }

  CopyVector(Vector(pTS, maxIndex), Vector(pCB,prototype), VectorSize(pCB));
  } /* FindMedoidQuick */

/*-------------------------------------------------------------------*/

void CATGenerateMedoidsQuick(PARTITIONING *pP,
                             TRAININGSET *pTS,
                             CATCODEBOOK *pCCB)
  {
  CATGenerateDistributions(pP, pTS, pCCB);
  int c;
  int size = BookSize(pCCB->pCodebook);

  for (c = 0; c < size; c++)
    {
    FindMedoidQuick(pP, pTS, pCCB, c);
    }
  } /* CATGenerateMedoidsQuick */

/*-------------------------------------------------------------------*/

void FindMode(CATCODEBOOK *pMCB, int prototype)
  {
  CODEBOOK *pCB = pMCB->pCodebook;
  int i, j, ins;
  int MaxInstances = -1, MaxCategory = -1;

  for (i=0; i<VectorSize(pCB); i++)
    {
    MaxInstances = -1;

    for (j=0; j<Categories(pMCB, i); j++)
      {
      ins = CategoryInstances(pMCB, prototype, i, j);

      if (ins > MaxInstances)
        {
        MaxCategory = j;
        MaxInstances = ins;
        }
      }

    VectorScalar(pCB, prototype, i) = MaxCategory;
    }
  } /* FindMode */

/*-------------------------------------------------------------------*/

void CATGenerateModes(PARTITIONING *pP,
                      TRAININGSET *pTS,
                      CATCODEBOOK *pCCB)
  {
  CATGenerateDistributions(pP, pTS, pCCB);
  int c;
  int size = BookSize(pCCB->pCodebook);

  for (c = 0; c < size; c++)
    {
    FindMode(pCCB, c);
    }
  } /* CATGenerateModes */

/*-------------------------------------------------------------------*/

void CATGenerateOptimalCodebook(PARTITIONING *pP,
                                TRAININGSET *pTS,
                                CATCODEBOOK *pCCB,
                                CATCENTROIDTYPE centroidType)
  {
  switch (centroidType)
    {
    case DISTRIBUTIONREPR:
      {
      CATGenerateDistributions(pP, pTS, pCCB);
      break;
      }
    case MEDOIDREPR:
      {
      CATGenerateMedoids(pP, pTS, pCCB, 0);
      break;
      }
    case MODEREPR:
      {
      CATGenerateModes(pP, pTS, pCCB);
      break;
      }
    case STMEDOIDREPR:
      {
      CATGenerateMedoidsQuick(pP, pTS, pCCB);
      break;
      }
    default:
      {
      ErrorMessage("ERROR: unsupported representative type %d\n",
                   centroidType);
      ExitProcessing(-1);
      }
    }
  } /* CATGenerateOptimalCodebook */

/*-------------------------------------------------------------------*/

int CATGenerateOptimalPartition(CATCODEBOOK *pCCB,
                                TRAININGSET *pTS,
                                PARTITIONING *pP,
                                CATDISTANCETYPE disttype)
  {
  int i;
  int nearest;
  int changes = 0;

  for (i = 0; i < BookSize(pTS); i++)
    {
    nearest = CATFindNearestVector(pCCB, pP, pTS, i, disttype);

    if (nearest != Map(pP, i))
      {
      if (CCFreq(pP, Map(pP, i)) != 1)
        {
        ChangePartition(pTS, pP, nearest, i);
        changes++;
        }
      }
    }

  return changes;
  } /* CATGenerateOptimalPartition */

/*-------------------------------------------------------------------*/

double CATClusterEntropy(CATCODEBOOK* pCCB, int i)
  {
  CODEBOOK* pCB = pCCB->pCodebook;
  double error = 0.0, probability;
  int j, k, count;

  for (j = 0; j < VectorSize(pCB); j++)
    {
    for (k = 0; k < Categories(pCCB, j); k++)
      {
      if (CategoryInstances(pCCB, i, j, k) != 0)
        {
        count = Count(pCCB, i);
        probability = (double)CategoryInstances(pCCB, i, j, k) /
                      (double)count;
        error -= probability * log(probability);
        }
      }
    }

  return error * Count(pCCB, i);
  } /* CATClusterEntropy */

/*-------------------------------------------------------------------*/

double CATDatasetEntropy(CATCODEBOOK* pCCB)
  {
  CODEBOOK* pCB = pCCB->pCodebook;
  double error = 0.0, probability;
  int i, j, k, count, categories;

  for (j = 0; j < VectorSize(pCB); j++)
    {
    for (k = 0; k < Categories(pCCB, j); k++)
      {
      categories = 0;
      count = 0;

      for (i = 0; i < BookSize(pCB); i++)
        {
        categories += CategoryInstances(pCCB, i, j, k);
        count      += Count(pCCB, i);
        }

      if (categories != 0)
        {
        probability = (double)categories /
                      (double)count;
        error -= probability * log(probability);
        }
      }
    }

  return error;
  } /* CATDatasetEntropy */

/*-------------------------------------------------------------------*/

double CATExpectedEntropy(CATCODEBOOK* pCCB, TRAININGSET* pTS)
  {
  CODEBOOK* pCB = pCCB->pCodebook;
  double ee = 0;
  int i;

  for (i = 0; i < BookSize(pCB); i++)
    {
    ee += CATClusterEntropy(pCCB, i);
    }

  return ee / BookSize(pTS);
  } /* CATExpectedEntropy */

/*-------------------------------------------------------------------*/

int PerformCKM(PARTITIONING *pP,
               CATCODEBOOK *pCCB,
               TRAININGSET *pTS,
               int kmIter,
               int quietLevel,
               CATDISTANCETYPE distanceMetric,
               CATCENTROIDTYPE centroidType,
               CATINITMETHOD initMethod)
  {
  CATPrint(quietLevel, INFOMESSAGE, "PerformCKM started\n");

  int i, j, changes;

  CATClearDivisions(pCCB);
  CATGenerateInitialSolution(pP, pCCB, pTS, initMethod);
  CATGenerateOptimalCodebook(pP, pTS, pCCB, centroidType);

  CATPrint(quietLevel, INFOMESSAGE, "InitialSolution created\n");

  for (i = 0; i < kmIter || kmIter == 0; i++)
    {
    changes = CATGenerateOptimalPartition(pCCB, pTS, pP, distanceMetric);
    CATPrintIterationInfo(quietLevel, i + 1, changes);

    if (changes == 0)
      {
      break;
      }

    CATGenerateOptimalCodebook(pP, pTS, pCCB, centroidType);

    for (j = 0; j < BookSize(pCCB->pCodebook); j++)
      {
      CATPrintV(quietLevel, DEBUGMESSAGE, "Prototype %d ( %d ):\n",
                j, Count(pCCB, j));
      }
    }

  CATPrint(quietLevel, INFOMESSAGE, "PerformCKM completed\n");

  return 0;
  } /* PerformCKM */

/*-------------------------------------------------------------------*/

void CATPrintIterationInfo(int quietLevel, int iteration, int changes)
  {
  CATPrintV(quietLevel, INFOMESSAGE, "Iteration %d, ", iteration);
  CATPrintV(quietLevel, INFOMESSAGE, "changes %d\n", changes);
  } /* PrintIterationInfo */

/*-------------------------------------------------------------------*/

int UTMIndex(UTM* pUTM, int x, int y)
  {
  int N = pUTM->N;
  int temp;

  if (x < y)
    {
    temp = x;
    x = y;
    y = temp;
    }

  return y * (N + N - y - 1) / 2 + x - y;
  } /* UTMIndex */

/*-------------------------------------------------------------------*/

void UTMCreate(UTM* pUTM, int N)
  {
  pUTM->N = N;
  pUTM->Size = (N * N - N) / 2;
  pUTM->Data = (void**) allocate(sizeof(void*) * (pUTM->Size + 1)); // check +1
  } /* UTMCreate */

/*-------------------------------------------------------------------*/

void* UTMGet(UTM* pUTM, int x, int y)
  {
  int index = UTMIndex(pUTM, x, y);
  return pUTM->Data[index];
  } /* UTMGet */

/*-------------------------------------------------------------------*/

void UTMSet(UTM* pUTM, int x, int y, void* data)
  {
  int index = UTMIndex(pUTM, x, y);
  pUTM->Data[index] = data;
  } /* UTMSet */

/*-------------------------------------------------------------------*/

void UTMFree(UTM* pUTM)
  {
  deallocate(pUTM->Data);
  } /* UTMFree */

/*-------------------------------------------------------------------*/

void CATPrintCluster(CATCODEBOOK *pCCB, int cluster)
  {
  int a, c;
  int attributes = VectorSize(pCCB->pCodebook);
  int categories;
  int instances;
  int first;
  char temp[10];

  for (a = 0; a < attributes; a++)
    {
    PrintMessage("Attr. %-5d|", a);
    categories = Categories(pCCB, a);
    first = 1;

    for (c = 0; c < categories; c++)
      {
      instances = CategoryInstances(pCCB, cluster, a, c);

      if (instances != 0)
        {
        if (!first)
          {
          PrintMessage("| ");
          }
        else
          {
          first = 0;
          }

        sprintf(temp, "%d:", c);
        PrintMessage("%5s %-5d", temp, instances);
        }
      }

    PrintMessage("\n");
    }
  } /* CATPrintCluster */

/*-------------------------------------------------------------------*/

void CATPrintCodebook(CATCODEBOOK *pCCB)
  {
  int i;
  int size = BookSize(pCCB->pCodebook);

  for (i = 0; i < size; i++)
    {
    PrintMessage("Representative %d\n", i);
    CATPrintCluster(pCCB, i);
    PrintMessage("\n");
    }
  } /* CATPrintCodebook */

/*-------------------------------------------------------------------*/

void CATPermutation(int* permutation, int size, int p)
  {
  int i;
  int index, temp;

  for (i = 0; i < size; i++)
    {
    permutation[i] = i;
    }

  for (i = 2; i <= size; i++)
    {
    index = (p % i);
    temp = permutation[i - 1];
    permutation[i - 1] = permutation[index];
    permutation[index] = temp;

    p = p / i;
    }
  } /* Permutation */

/*-------------------------------------------------------------------*/

int CATFactorial(int i)
  {
  return (i <= 1) ? 1 : i * CATFactorial(i - 1);
  } /* Factorial */

/*-------------------------------------------------------------------*/

void InitRandom()
  {
  struct timeb tp;
  ftime(&tp);
  long seed = tp.time + 1000000L * tp.millitm;
  initrandom(seed);
  } /* InitRandom */

/*-------------------------------------------------------------------*/

