/*
$Revision: 1.11 $
$Date: 2009/12/01 11:59:48 $
$Author: apollan $
$Name:  $
$Id: cbrock.c,v 1.11 2009/12/01 11:59:48 apollan Exp $
*/

/*-------------------------------------------------------------------*/
/* ROCK.C     Antti Pollanen                                         */
/*                                                                   */
/* Console interface ROCK -algorithm                                 */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/


#define  ProgName       "cbrock"
#define  VersionNumber  "Version 0.51"
#define  LastUpdated    "01.12.2009" /* AP */
#define  MAXFILENAME    64
#define  FACTFILE       "cbrock.fac"

/*-------------------------------------------------------------------*/

#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"

#include "heap.h"
#include "ctk.h"
#include "rock.h"

#include "parametr.c"


/* ========================== PROTOTYPES =========================== */

void PrintInitialData(int quietLevel,
                      char *TSName,
                      char *PAName,
                      int k,
                      int maxK,
                      double theta);

void PrintInfo(void);

int PerformROCKMultipleResults(PARTITIONING* pP,
                               TRAININGSET* pTS,
                               int k,
                               int maxK,
                               double theta,
                               int quietLevel,
                               int* outlierCluster,
                               char* PAName);

int main(int argc, char* argv[]);


/* ========================== FUNCTIONS ============================ */

void PrintInitialData(int quietLevel,
                      char *TSName,
                      char *PAName,
                      int k,
                      int maxK,
                      double theta)
  {
  if (quietLevel >= INFOMESSAGE)
    {
    PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
    PrintMessage("Training Set\t\t = %s\n", TSName);
    PrintMessage("Result Partitioning\t = %s\n", PAName);

    if (maxK > k)
      {
      PrintMessage("Codebook size\t\t = %d - %d\n", k, maxK);
      }
    else
      {
      PrintMessage("Codebook size\t\t = %d\n", k);
      }

    PrintMessage("Theta\t\t\t = %f\n", theta);
    }
  }  /* PrintInitialData() */

/*-------------------------------------------------------------------*/

void PrintInfo(void)
  {
  PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
  PrintMessage("Console interface for the ROCK-algorithm.\n");
  PrintMessage("Usage: %s [%coption] ", ProgName, OPTION_SYMBOL);
  PrintMessage("<training set> <result partitioning>\n");
  PrintMessage("For example: %s bridge tmp\n", ProgName);
  PrintMessage("\n  Options:\n");
  PrintOptions();
  PrintMessage("\n");
  }  /* PrintInfo() */


void WriteOutlierCluster(TRAININGSET* pTS, PARTITIONING* pP, int outlierCluster)
  {
  int i, j;
  FILE* file = FileOpen("outliers", OUTPUT, 1);
  fprintf(file, "Outlier cluster: %d\n", outlierCluster+1);


  i = FirstVector(pP, outlierCluster);
  fprintf(file, "Vectors (%d):\n", CCFreq(pP, outlierCluster));

  while(!EndOfPartition(i))
    {
    fprintf(file, "%d:\t", i);

    for (j = 0; j < VectorSize(pTS); j++)
      {
      fprintf(file, "%d\t", VectorScalar(pTS, i, j));
      }

    fprintf(file, "\n");
    i = NextVector(pP, i);
    }

  fclose(file);
  }

/*-------------------------------------------------------------------*/

int PerformROCKMultipleResults(PARTITIONING* pP,
                               TRAININGSET* pTS,
                               int k,
                               int maxK,
                               double theta,
                               int quietLevel,
                               int* outlierCluster,
                               char* PAName)
  {
  char    RealPAName[MAXFILENAME + 10];
  int i, error = 0;
  int BookSize = BookSize(pTS);
  ROCKDATA data;

  CATPrint(quietLevel, INFOMESSAGE, "Clustering...\n");
  BookSize = ROCKInit(pP, pTS, &data, theta, quietLevel);

  if (BookSize != BookSize(pTS))
    {
    *outlierCluster = 0;
    }

  if (BookSize <= maxK)
    {
    sprintf(RealPAName, "%.*s%d.pa", strlen(PAName) - 3, PAName, BookSize(pTS));
    WritePartitioning(RealPAName, pP, pTS, 1);
    }

  for (i = BookSize; i > k && !error; i--)
    {
    error = error || ROCKIteration(pP, pTS, &data, quietLevel);

    if (i - 1 <= maxK)
      {
      sprintf(RealPAName, "%.*s%d.pa", strlen(PAName) - 3, PAName, i-1);
      WritePartitioning(RealPAName, pP, pTS, 1);
      }
    }

  ROCKFree(&data, quietLevel);
  CATPrint(quietLevel, INFOMESSAGE, "Clustering completed.\n");

  return error;
  } /* PerformROCK */

/*-------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
  TRAININGSET   TS;
  PARTITIONING  P;
  char          TSName[MAXFILENAME] = {'\0'};
  char          PAName[MAXFILENAME] = {'\0'};
  ParameterInfo paraminfo[2] = { { TSName,  FormatNameTS, 0, INFILE },
                                 { PAName,  FormatNamePA, 0, OUTFILE } };
  double ScaledTheta;
  int result, i;
  int outlierCluster = -1;

  ParseParameters(argc, argv, 2, paraminfo);

  /* training set doesn't exist */
  if (!ExistFile(TSName))
    {
    ErrorMessage("\nERROR: File doesn't exist: %s\n\n", TSName);
    ExitProcessing(FATAL_ERROR);
    }

  /* result codebook exists and we are told not to overwrite */
  if (ExistFile(PAName) && !Value(OverWrite))
    {
    ErrorMessage("\nERROR: File already exists: %s\n\n", PAName);
    ExitProcessing(FATAL_ERROR);
    }

  ScaledTheta = (double)Value(Theta) / 1000.0;

  ReadTrainingSet(TSName, &TS);

  /* codebook cannot contain more vectors than training set */
  if (BookSize(&TS) < Value(Clusters))
    {
    ErrorMessage("\nERROR: Number of vectors in training set ");
    ErrorMessage("(%d) < number of clusters ", BookSize(&TS));
    ErrorMessage("(%d)!\n\n", Value(Clusters));
    FreeCodebook(&TS);
    ExitProcessing(FATAL_ERROR);
    }

  CreateNewPartitioning(&P, &TS, BookSize(&TS) + 1);

  PrintInitialData(Value(QuietLevel),
                  TSName,
                  PAName,
                  Value(Clusters),
                  Value(MaxClusters),
                  ScaledTheta);

  result = PerformROCKMultipleResults(&P,
                                      &TS,
                                      Value(Clusters),
                                      Value(MaxClusters),
                                      ScaledTheta,
                                      Value(QuietLevel),
                                      &outlierCluster,
                                      PAName);

  if (result)
    {
    ErrorMessage("ERROR: Clustering failed!\n");

    if (result == ERROR_IMPOSSIBLE_NUMBER_OF_CLUSTERS)
      {
      ErrorMessage("There were not enough cluster pairs.\n");
      ErrorMessage("Try smaller Theta value.\n");
      }

    FreeCodebook(&TS);
    FreePartitioning(&P);
    ExitProcessing(FATAL_ERROR);
    }

  if (outlierCluster < 0)
    {
    for (i = 0; i < BookSize(&TS); i++)
      {
      ChangePartition(&TS, &P, Map(&P, i) - 1, i);
      }
    DecreaseNumberOfPartitions(&P, PartitionCount(&P) - 1);
    }
  else
    {
    PrintMessage("Outlier cluster: %d\n", outlierCluster+1);
    }

  WritePartitioning(PAName, &P, &TS, Value(OverWrite));

  if (Value(WriteOutlierCluster))
    {
    WriteOutlierCluster(&TS, &P, outlierCluster);
    }

  FreePartitioning(&P);
  FreeCodebook(&TS);

  checkmemory();
  return 0;
}
