/*
 $Revision: 1.13 $
 $Date: 2009/12/03 13:46:01 $
 $Author: apollan $
 $Name:  $
 $Id: rock.c,v 1.13 2009/12/03 13:46:01 apollan Exp $
 */

/*-------------------------------------------------------------------*/
/* ROCK.C       Antti Pollanen                                       */
/*                                                                   */
/* ROCK -algorithm                                                   */
/*                                                                   */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    CCB       categorical data codebook                            */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "rock"
#define  VersionNumber  "Version 0.51"
#define  LastUpdated    "01.12.2009" /* AP */

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* AP   0.51  Multiple partitions bug fix and merge bug fix          */
/* AP   0.50  Many partitions with single run.                       */
/* AP   0.43  Link calculations on merging fixed.                    */
/* AP   0.42  Outlier handling changed: printing and file            */
/* AP   0.41  Indexing bug fix: 1-to-(n-1) -> 1-to-n                 */
/* AP   0.40  Outlier handling implemented (0-cluster)               */
/* AP   0.30  Error in MergeClusters function fixed                  */
/* AP   0.20  Better structure and outlier handling                  */
/* AP   0.12  Stardardization for coding and commenting              */
/* AP   0.11  Code formatting.                                       */
/* AP   0.10  Method fixed by using UTM                              */
/* AP   0.01  Initial version                                        */
/*                                                                   */
/* ----------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>

#include "parametr.h"
#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "math.h"
#include "heap.h"
#include "stack.h"

#include "ctk.h"
#include "rock.h"

void MergeClusters(PARTITIONING* pP, TRAININGSET* pTS, ROCKDATA* data,
    int quietLevel, int c1, int c2);

void RenameCluster(PARTITIONING* pP, TRAININGSET* pTS, ROCKDATA* data,
    int quietLevel, int c2);

/* ========================== FUNCTIONS ============================ */

int IsInHeap(HEAP* pHeap, CLUSTERPAIR* pPair)
  {
  if (pPair->Index > 0 && pPair->Index <= Heap_size(pHeap))
    {
    return 1;
    }

  return 0;
  } /* IsInHeap */

/*-------------------------------------------------------------------*/

double Similarity(TRAININGSET* pTS, int v1, int v2)
  {
  int same = 0;
  int total = 0;
  int i;

  for (i = 0; i < VectorSize(pTS); i++)
    {
    if (VectorScalar(pTS, v1, i) == VectorScalar(pTS, v2, i))
      {
      same++;
      }
    else
      {
      total++;
      }

    total++;
    }

  return (double) same / (double) total;
  } /* Similarity */

/*-------------------------------------------------------------------*/

void ComputeLinks(TRAININGSET* pTS, int* Links, double theta)
  {
  int size = BookSize(pTS);
  int* nbrlist = (int*) allocate(size * sizeof(int));
  int i, j, l, listSize;

  for (i = 0; i < size; i++)
    {
    listSize = 0;

    // Compute neighbour list.
    for (j = 0; j < size; j++)
      {
      if (i != j && Similarity(pTS, i, j) > theta)
        {
        nbrlist[listSize] = j;
        listSize++;
        }
      }

    // Compute actual links
    for (j = 0; j < listSize; j++)
      {
      for (l = j + 1; l < listSize; l++)
        {
        Links[nbrlist[j]*size + nbrlist[l]]++;
        }
      }
    }

  deallocate(nbrlist);
  } /* ComputeLinks */

/*-------------------------------------------------------------------*/

int HandleOutliers(TRAININGSET* pTS, PARTITIONING* pP, int* Links, int bookSize, int quietLevel)
  {
  int outlier;
  int outliers = 0;
  int i, j;

  // Generate initial partitioning
  for (i = 0; i < bookSize; i++)
    {
    outlier = 1;

    for (j = 0; j < bookSize; j++)
      {
      if (i < j && Links[i*bookSize + j] > 0)
        {
        outlier = 0;
        }
      else if (i > j && Links[j*bookSize + i] > 0)
        {
        outlier = 0;
        }
      }

    if (outlier)
      {
      outliers++;
      ChangePartition(pTS, pP, 0, i);
      }
    else
      {
      ChangePartition(pTS, pP, i - outliers + 1, i);
      }
    }

  bookSize = bookSize - outliers;

  CATPrintV(quietLevel, DEBUGMESSAGE, "Outliers  %d\n", outliers);
  CATPrintV(quietLevel, DEBUGMESSAGE, "BookSize  %d\n", bookSize);

  if (outliers > 0)
    {
    DecreaseNumberOfPartitions(pP, PartitionCount(pP) - outliers);
    }

  return bookSize;
  } /* HandleOutliers */

/*-------------------------------------------------------------------*/

double f(double theta)
  {
  return (1.0 - theta) / (1.0 + theta);
  } /* f */

/*-------------------------------------------------------------------*/

double Goodness(PARTITIONING* pP, UTM* pUTM, int i, int j, double theta)
  {
  double p = 1.0 + 2.0 * f(theta);
  int c1 = CCFreq(pP, i);
  int c2 = CCFreq(pP, j);
  double goodness;

  CLUSTERPAIR* pPair = (CLUSTERPAIR*) UTMGet(pUTM, i, j);
  goodness = (double) (pPair->Links) /
             (double) (pow(c1 + c2, p) - (pow(c1, p) + pow(c2, p)));

  return goodness;
  } /* Goodness */

/*-------------------------------------------------------------------*/

int ClusterCompare(void* elem1, void* elem2, void* info)
  {
  CLUSTERPAIR* c1 = (CLUSTERPAIR*) elem1;
  CLUSTERPAIR* c2 = (CLUSTERPAIR*) elem2;

  if (c1->Cost < c2->Cost)
    {
    return 1;
    }
  else if (c2->Cost < c1->Cost)
    {
    return -1;
    }

  return EQUAL;
  } /* ClusterCompare */

/*-------------------------------------------------------------------*/

void InitializePairs(PARTITIONING* pP, ROCKDATA* data, CLUSTERPAIR* Pairs,
  int* Links, int bookSize, int size, int quietLevel)
  {
  UTM *pUtm = &(data->utm);
  int Index = 0;
  int i, j;
  CLUSTERPAIR *pPair;

  for (i = 1; i <= size; i++)
    {
    for (j = i + 1; j <= size; j++)
      {
      pPair = &(Pairs[Index]);
      pPair->Cluster1 = i;
      pPair->Cluster2 = j;
      pPair->Links = Links[FirstVector(pP, i)*bookSize + FirstVector(pP, j)];
      UTMSet(pUtm, i, j, pPair);
      Index++;
      }
    }
  }

/*-------------------------------------------------------------------*/

int BuildGlobalHeap(PARTITIONING* pP, TRAININGSET* pTS, ROCKDATA* data,
    int bookSize, int quietLevel)
  {
  UTM*  pUtm  = &(data->utm);
  HEAP* pHeap = &(data->GlobalHeap);
  double theta = data->Theta;

  int i, j;
  int heapSize = 0;
  CLUSTERPAIR* pPair;

  for (i = 1; i <= bookSize; i++)
    {
    for (j = i + 1; j <= bookSize; j++)
      {
      pPair = (CLUSTERPAIR*) UTMGet(pUtm, i, j);

      if ((pPair->Links) > 0)
        {
        heapSize++;
        }
      }
    }

  CATPrintV(quietLevel, DEBUGMESSAGE, "Heap size %d\n", heapSize);
  Heap_init(pHeap, heapSize, ClusterCompare);

  for (i = 1; i <= bookSize; i++)
    {
    for (j = i + 1; j <= bookSize; j++)
      {
      pPair = (CLUSTERPAIR*) UTMGet(pUtm, i, j);

      if ((pPair->Links) > 0)
        {
        pPair->Cost = Goodness(pP, pUtm, i, j, theta);
        Heap_insert(pHeap, pPair, NULL, &(pPair->Index));
        }
      }
    }

  return bookSize;
  } /* BuildGlobalHeap */

/*-------------------------------------------------------------------*/

int ROCKInit(PARTITIONING* pP,
             TRAININGSET* pTS,
             ROCKDATA* data,
             double theta,
             int quietLevel)
  {
  CATPrint(quietLevel, INFOMESSAGE, "Initializing...\n");
  int bookSize = BookSize(pTS), size;

  data->Theta = theta;
  UTM* pUtm = &(data->utm);
  int* Links;
  CLUSTERPAIR* Pairs;

  InitRandom();

  CATPrint(quietLevel, INFOMESSAGE, "Computing links ...\n");
  Links = (int*) allocate(sizeof(int) * (bookSize*bookSize));
  ComputeLinks(pTS, Links, theta);

  CATPrint(quietLevel, INFOMESSAGE, "Handling outliers ...\n");
  size = HandleOutliers(pTS, pP, Links, bookSize, quietLevel);

  CATPrint(quietLevel, INFOMESSAGE, "Initializing Pairs ...\n");
  UTMCreate(pUtm, size+1);
  data->Pairs = (CLUSTERPAIR*) allocate(sizeof(CLUSTERPAIR) * (pUtm->Size));
  Pairs = data->Pairs;
  InitializePairs(pP, data, Pairs, Links, bookSize, size, quietLevel);

  CATPrint(quietLevel, INFOMESSAGE, "Building Global heap ... \n");
  BuildGlobalHeap(pP, pTS, data, size, quietLevel);

  deallocate(Links);
  data->x = size;
  CATPrint(quietLevel, INFOMESSAGE, "Initialization completed. \n");

  return size;
  } /* ROCKInit */

/*-------------------------------------------------------------------*/

void RenameCluster(PARTITIONING* pP, TRAININGSET* pTS, ROCKDATA* data,
    int quietLevel, int c2)
  {
  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->GlobalHeap);
  CLUSTERPAIR* Pairs = data->Pairs;
  int x = data->x;

  int i;
  int last = x;
  int renamed = 0;

  // Rename last cluster to c2.
  if (c2 != last)
    {
    CATPrintV(quietLevel, INFOMESSAGE, "Renaming %d to %d.\n", last, c2);

    for (i = 1; i < last; i++)
      {
      if (c2 != i)
        {
        // Pair (c2, last) is removed already.
        CLUSTERPAIR* MovePair = UTMGet(pUtm, i, last);

        if (MovePair->Cluster1 != i && MovePair->Cluster2 != i)
          {
          CATPrintV(quietLevel, ERRORMESSAGE,
                    "ERROR: renaming wrong pair (%d, %d).\n",
                    MovePair->Cluster1, MovePair->Cluster2);
          }

        renamed++;

        if (c2 < i)
          {
          MovePair->Cluster1 = c2;
          MovePair->Cluster2 = i;
          UTMSet(pUtm, c2, i, MovePair);
          }
        else
          {
          MovePair->Cluster1 = i;
          MovePair->Cluster2 = c2;
          UTMSet(pUtm, i, c2, MovePair);
          }
        }
      }
    }

  CATPrintV(quietLevel, DEBUGMESSAGE, "Renamed %d elements.\n", renamed);

  for (i = 0; i < pUtm->Size; i++)
    {
    if (IsInHeap(pHeap, &(Pairs[i])) &&
        (Pairs[i].Cluster1 >= last || Pairs[i].Cluster2 >= last))
      {
      CATPrintV(quietLevel, ERRORMESSAGE,
                "ERROR: Pair (%d, %d) found!\n",
                Pairs[i].Cluster1, Pairs[i].Cluster2);
      }
    }
  }

/*-------------------------------------------------------------------*/

void MergeClusters(PARTITIONING* pP, TRAININGSET* pTS, ROCKDATA* data,
    int quietLevel, int c1, int c2)
  {
  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->GlobalHeap);
  double theta = data->Theta;
  int x = data->x;

  int i;
  int removed = 0;
  CLUSTERPAIR* OldPair;
  CLUSTERPAIR* pPair;

  // Remove cluster c2.
  for (i = 1; i <= x; i++) // = ?
    {
    // Pair(x, x) doesn't exist and Pair(c1, c2) is already removed.
    if (i != c2 && i != c1)
      {
      OldPair = UTMGet(pUtm, c2, i);
      pPair = UTMGet(pUtm, c1, i);

      if (IsInHeap(pHeap, OldPair))
        {
        if (IsInHeap(pHeap, pPair))
          {
          pPair->Links += OldPair->Links;
          }
        else
          {
          pPair->Links = OldPair->Links;
          pPair->Cost = Goodness(pP, pUtm, c1, i, theta);
          Heap_insert(pHeap, pPair, NULL, &(pPair->Index));
          }

        Heap_remove(pHeap, OldPair->Index, NULL);

        if (IsInHeap(pHeap, OldPair))
          {
          CATPrintV(quietLevel, ERRORMESSAGE,
                    "ERROR: heap is broken (index %d)!\n",
                    OldPair->Index);
          }

        if (OldPair->Cluster1 != c2 && OldPair->Cluster2 != c2)
          {
          CATPrintV(quietLevel, ERRORMESSAGE,
                    "ERROR: removing wrong pair (%d, %d).\n",
                    OldPair->Cluster1, OldPair->Cluster2);
          }

        removed++;
        }
      else
        {
        // PrintMessage("Pair (%d, %d) NOT REMOVED!\n", c2, i);
        // This is ok, because in ROCK heap is not "full".
        // CATPrintV(quietLevel, ERRORMESSAGE,
        //           "Cluster pair %d, %d doesn't exist.\n", c2, i);
        }
      }
    }

  CATPrintV(quietLevel, DEBUGMESSAGE, "Removed %d pairs.\n", removed);
  JoinPartitions(pTS, pP, c1, c2);
  }

/*-------------------------------------------------------------------*/

void UpdateMergeCosts(PARTITIONING* pP, TRAININGSET* pTS, ROCKDATA* data,
    int quietLevel, int c1)
  {
  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->GlobalHeap);
  double theta = data->Theta;
  int x = data->x;

  CLUSTERPAIR* OldPair;
  int i;

  // Update merge costs, cluster x is removed all ready.
  for (i = 1; i < x; i++)
    {
    if (i != c1)
      {
      OldPair = UTMGet(pUtm, c1, i);

      if (IsInHeap(pHeap, OldPair))
        {
        OldPair->Cost = Goodness(pP, pUtm, c1, i, theta);
        Heap_update(pHeap, OldPair->Index, NULL);
        }
      else
        {
        // This is ok, because in ROCK heap is not full.
        // CATPrintV(quietLevel, ERRORMESSAGE,
        //           "Cluster pair %d, %d doesn't exist.\n", c2, i);
        }
      }
    }
  }

/*-------------------------------------------------------------------*/

CLUSTERPAIR* SelectPair(HEAP* pHeap)
  {
  CLUSTERPAIR* pPair;
  CLUSTERPAIR* pFirst;
  CLUSTERPAIR* pSelected = 0;
  STACK* stack = S_make();
  int size = 0;

  pFirst = (CLUSTERPAIR*) Heap_removeroot(pHeap, NULL);
  S_push(stack, pFirst);
  size++;
  int selected;
  int tie = 1;

  while (tie && !Heap_isempty(pHeap))
    {
    pPair = (CLUSTERPAIR*) Heap_removeroot(pHeap, NULL);

    if (fabs(pPair->Cost - pFirst->Cost) < EPSILON)
      {
      S_push(stack, pPair);
      size++;
      }
    else
      {
      Heap_insert(pHeap, pPair, NULL, &(pPair->Index));
      tie = 0;
      }
    }

  selected = irand(0, size-1);

  while (!S_empty(stack))
    {
    pPair = (CLUSTERPAIR*)S_pop(stack);
    size--;

    if (size == selected)
      {
      pSelected = pPair;
      }
    else
      {
      Heap_insert(pHeap, pPair, NULL, &(pPair->Index));
      }
    }

  S_free(stack);

  return pSelected;
  } /* SelectPair */

/*-------------------------------------------------------------------*/

int ROCKIteration(PARTITIONING* pP, TRAININGSET* pTS, ROCKDATA* data,
    int quietLevel)
  {
  HEAP* pHeap = &(data->GlobalHeap);
  int x = data->x;

  CLUSTERPAIR* pPair;
  int c1, c2;

  CATPrintV(quietLevel, INFOMESSAGE, "--- Iteration starts ---\n", x);

  if (Heap_isempty(pHeap))
    {
    return ERROR_IMPOSSIBLE_NUMBER_OF_CLUSTERS;
    }

  pPair = SelectPair(pHeap);
  c1 = pPair->Cluster1;
  c2 = pPair->Cluster2;

  CATPrintV(quietLevel, INFOMESSAGE, "Merging %d and %d\n", c1, c2);
  MergeClusters(pP, pTS, data, quietLevel, c1, c2);

  RenameCluster(pP, pTS, data, quietLevel, c2);

  CATPrint(quietLevel, INFOMESSAGE, "Updating merge costs\n");
  UpdateMergeCosts(pP, pTS, data, quietLevel, c1);

  data->x = --x;
  CATPrintV(quietLevel, INFOMESSAGE, "----- Codebook size %d -----\n", x);
  return 0;
  } /* ROCKIteration */

/*-------------------------------------------------------------------*/

int ROCKFree(ROCKDATA* data, int quietLevel)
  {
  UTM* pUtm = &(data->utm);
  HEAP* pHeap = &(data->GlobalHeap);
  CLUSTERPAIR* Pairs = data->Pairs;

  UTMFree(pUtm);
  Heap_free(pHeap);
  deallocate(Pairs);

  return 0;
  } /* ROCKFree */

/*-------------------------------------------------------------------*/

int PerformROCK(PARTITIONING* pP,
                TRAININGSET* pTS,
                int k,
                double theta,
                int quietLevel,
                int* outlierCluster)
  {
  int i, error = 0;
  int BookSize = BookSize(pTS);
  ROCKDATA data;

  CATPrint(quietLevel, INFOMESSAGE, "Clustering...\n");
  BookSize = ROCKInit(pP, pTS, &data, theta, quietLevel);

  if (BookSize != BookSize(pTS))
    {
    *outlierCluster = 0;
    }

  for (i = BookSize; i > k && !error; i--)
    {
    error = error || ROCKIteration(pP, pTS, &data, quietLevel);
    }

  ROCKFree(&data, quietLevel);
  CATPrint(quietLevel, INFOMESSAGE, "Clustering completed.\n");

  return error;
  } /* PerformROCK */

/*-------------------------------------------------------------------*/
